"""
Copyright 2018 Inmanta

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Contact: code@inmanta.com
"""

import asyncio
import logging
import os
import shutil
import sys
import time
import uuid
from asyncio import queues, subprocess
from collections.abc import Iterable, Iterator, Mapping, Sequence, Set
from dataclasses import dataclass
from datetime import datetime
from enum import Enum
from functools import reduce
from typing import Any, Optional, Union, assert_never, cast
from uuid import UUID

import asyncpg.connection

import inmanta.config
import inmanta.exceptions
import inmanta.server.services.environmentlistener
from inmanta import config as global_config
from inmanta import const, data
from inmanta import logging as inmanta_logging
from inmanta import tracing
from inmanta.agent import config as agent_cfg
from inmanta.config import Config, config_map_to_str, scheduler_log_config
from inmanta.const import AGENT_SCHEDULER_ID, UNDEPLOYABLE_NAMES, AgentAction, AgentStatus, AllAgentAction
from inmanta.data import APILIMIT, Environment, InvalidSort, model
from inmanta.data.model import DataBaseReport
from inmanta.protocol import encode_token, handle, methods, methods_v2
from inmanta.protocol.common import ReturnValue
from inmanta.protocol.exceptions import BadRequest, Conflict, Forbidden, NotFound, ShutdownInProgress
from inmanta.server import (
    SLICE_AGENT_MANAGER,
    SLICE_AUTOSTARTED_AGENT_MANAGER,
    SLICE_DATABASE,
    SLICE_ENVIRONMENT,
    SLICE_SERVER,
    SLICE_SESSION_MANAGER,
    SLICE_TRANSPORT,
)
from inmanta.server import config as opt
from inmanta.server import protocol
from inmanta.server.protocol import ReturnClient, ServerSlice, SessionListener, SessionManager
from inmanta.server.server import Server
from inmanta.server.services import environmentservice
from inmanta.types import Apireturn, ArgumentTypes, ResourceIdStr, ReturnTupple

from ..data.dataview import AgentView
from . import config as server_config
from .validate_filter import InvalidFilter

LOGGER = logging.getLogger(__name__)


"""
Model in server         On Agent

+---------------+        +----------+
|               |        |          |
|  ENVIRONMENT  |   +---->  PROC    |
|               |   |    |          |
+------+--------+   |    +----+-----+
       |            |         |
       |            |         |
+------v--------+   |    +----v-------------+
|               |   |    |                  |
|  AGENT        |   |    |   AGENT INSTANCE |
|               |   |    |                  |
+------+--------+   |    +------------------+
       |            |
       |            |
+------v--------+   |
|               |   |
|  SESSION      +---+
|               |
+---------------+

resource_action_update

dryrun_update

set_parameters
"""


class SessionActionType(str, Enum):
    REGISTER_SESSION = "register_session"
    EXPIRE_SESSION = "expire_session"
    SEEN_SESSION = "seen_session"


class SessionAction:
    """
    A session update to be executed by the AgentManager.
    """

    def __init__(
        self, action_type: SessionActionType, session: protocol.Session, endpoint_names_snapshot: set[str], timestamp: datetime
    ):
        self.action_type = action_type
        self.session = session
        self.endpoint_names_snapshot = endpoint_names_snapshot
        self.timestamp = timestamp


# Internal tuning constants

AUTO_STARTED_AGENT_WAIT = 5
# How long (in seconds) do we wait for autostarted agents. The wait time is reset any time a new instance comes online
AUTO_STARTED_AGENT_WAIT_LOG_INTERVAL = 1
# When waiting for an autostarted agent, how long (in seconds) do we wait before we log the wait status


class AgentManager(ServerSlice, SessionListener):
    """
    This class contains all server functionality related to the management of agents.
    Each logical agent managed by an instance of this class has at most one primary agent instance process associated with
    it. A subset of these processes are autostarted, those are managed by :py:class:`AutostartedAgentManager`.
    The server ignores all requests from non-primary agent instances. Therefore an agent without a primary is effectively
    paused as far as the server is concerned, though any rogue agent instances could still perform actions agent-side.

    Throughout this class the terms "logical agent" or sometimes just "agent" refer to a logical agent managed by an
    instance of this class. The terms "agent instance", "agent process" or just "process" refer to a concrete process
    running an agent instance, which might be the primary for a logical agent.
    """

    def __init__(self, closesessionsonstart: bool = True, fact_back_off: Optional[int] = None) -> None:
        super().__init__(SLICE_AGENT_MANAGER)

        if fact_back_off is None:
            fact_back_off = opt.server_fact_resource_block.get()

        # back-off timer for fact requests
        self._fact_resource_block: int = fact_back_off
        # per resource time of last fact request
        self._fact_resource_block_set: dict[str, float] = {}

        # session lock
        self.session_lock = asyncio.Lock()
        # all sessions
        self.sessions: dict[UUID, protocol.Session] = {}
        # live sessions: Sessions to agents which are primary and unpaused
        self.tid_endpoint_to_session: dict[tuple[UUID, str], protocol.Session] = {}
        # All endpoints associated with a sid
        self.endpoints_for_sid: dict[uuid.UUID, set[str]] = {}

        # This queue ensures that notifications from the SessionManager are processed in the same order
        # in which they arrive in the SessionManager, without blocking the SessionManager.
        self._session_listener_actions: queues.Queue[SessionAction] = queues.Queue()

        self.closesessionsonstart: bool = closesessionsonstart

    async def get_status(self) -> Mapping[str, ArgumentTypes | Mapping[str, ArgumentTypes]]:
        # The basic report

        out: dict[str, ArgumentTypes | Mapping[str, ArgumentTypes]] = {
            "resource_facts": len(self._fact_resource_block_set),
            "sessions": len(self.sessions),
        }

        # Try to get more info from scheduler, but make sure not to timeout
        schedulers = self.get_all_schedulers()
        deadline = 0.9 * Server.GET_SLICE_STATUS_TIMEOUT

        async def get_report(env: uuid.UUID, session: protocol.Session) -> tuple[uuid.UUID, DataBaseReport]:
            result = await asyncio.wait_for(session.client.get_db_status(), deadline)
            assert result.code == 200
            # Mypy can't help here, ....
            return (env, DataBaseReport(**result.result["data"]))

        # Get env name mapping in parallel with next call
        env_mapping = asyncio.create_task(
            asyncio.wait_for(Environment.get_list(details=False, is_marked_for_deletion=False), deadline)
        )
        # Get the reports of all schedulers
        results = await asyncio.gather(*[get_report(env, scheduler) for env, scheduler in schedulers], return_exceptions=True)
        try:
            # This can timeout, but not likely
            env_mapping_result = await env_mapping
            uuid_to_name = {env.id: env.name for env in env_mapping_result}
        except TimeoutError:
            # default to uuid's
            uuid_to_name = {}

        # Filter out timeouts and other errors
        good_reports = [x[1] for x in results if not isinstance(x, BaseException)]

        def short_report(report: DataBaseReport) -> Mapping[str, ArgumentTypes]:
            return {
                "connected": report.connected,
                "max_pool": report.max_pool,
                "open_connections": report.open_connections,
                "free_connections": report.free_connections,
                "pool_exhaustion_time": report.pool_exhaustion_time,
            }

        if good_reports:
            total: DataBaseReport = reduce(lambda x, y: x + y, good_reports)
            out["database"] = total.database
            out["host"] = total.host
            out["total"] = short_report(total)

        for result in results:
            if isinstance(result, BaseException):
                logging.debug("Failed to collect database status for scheduler", exc_info=True)
            else:
                the_uuid = result[0]
                # Default name to uuid
                name = uuid_to_name.get(the_uuid, str(the_uuid))
                out[name] = short_report(result[1])

        return out

    def get_dependencies(self) -> list[str]:
        return [SLICE_DATABASE, SLICE_SESSION_MANAGER]

    def get_depended_by(self) -> list[str]:
        return [SLICE_TRANSPORT]

    async def prestart(self, server: protocol.Server) -> None:
        await ServerSlice.prestart(self, server)
        autostarted_agent_manager = server.get_slice(SLICE_AUTOSTARTED_AGENT_MANAGER)
        assert isinstance(autostarted_agent_manager, AutostartedAgentManager)
        self._autostarted_agent_manager = autostarted_agent_manager
        presession = server.get_slice(SLICE_SESSION_MANAGER)
        assert isinstance(presession, SessionManager)
        presession.add_listener(self)

    async def start(self) -> None:
        await super().start()

        if self.closesessionsonstart:
            await self._expire_all_sessions_in_db()

        self.add_background_task(self._process_session_listener_actions())
        # Schedule cleanup agentprocess and agentinstance tables
        agent_process_purge_interval = opt.agent_process_purge_interval.get()
        if agent_process_purge_interval > 0:
            self.schedule(
                self._purge_agent_processes, interval=agent_process_purge_interval, initial_delay=0, cancel_on_stop=False
            )

    async def prestop(self) -> None:
        await super().prestop()

    async def stop(self) -> None:
        await super().stop()

    def get_all_schedulers(self) -> list[tuple[uuid.UUID, protocol.Session]]:
        # Linear scan, but every item should be a hit
        return [
            (env_id, session)
            for (env_id, agent_id), session in self.tid_endpoint_to_session.items()
            if agent_id == AGENT_SCHEDULER_ID
        ]

    async def halt_agents(self, env: data.Environment, connection: Optional[asyncpg.connection.Connection] = None) -> None:
        """
        Halts all agents for an environment. Persists prior paused state. Also halts the scheduler "agent"
        """
        await data.Agent.persist_on_halt(env.id, connection=connection)
        await self._pause_agent(env, connection=connection)  # excludes scheduler
        await self._pause_agent(env, endpoint=const.AGENT_SCHEDULER_ID, connection=connection)

    async def resume_agents(self, env: data.Environment, connection: Optional[asyncpg.connection.Connection] = None) -> None:
        """
        Resumes after halting. Unpauses all agents that had been paused by halting, then restarts the scheduler.
        """
        to_unpause: list[str] = await data.Agent.persist_on_resume(env.id, connection=connection)
        await asyncio.gather(*[self._unpause_agent(env, agent, connection=connection) for agent in to_unpause])
        await self._unpause_agent(env, endpoint=const.AGENT_SCHEDULER_ID, connection=connection)

    @handle(methods_v2.all_agents_action, env="tid")
    async def all_agents_action(self, env: data.Environment, action: AllAgentAction) -> None:
        if env.halted and action in {AllAgentAction.pause, AllAgentAction.unpause}:
            raise Forbidden("Can not pause or unpause agents when the environment has been halted.")
        if not env.halted and action in {AllAgentAction.keep_paused_on_resume, AllAgentAction.unpause_on_resume}:
            raise Forbidden("Cannot set on_resume state of agents when the environment is not halted.")
        match action:
            case AllAgentAction.pause:
                await self._pause_agent(env=env)
            case AllAgentAction.unpause:
                await self._unpause_agent(env=env)
            case AllAgentAction.keep_paused_on_resume:
                await self._set_unpause_on_resume(env, should_be_unpaused_on_resume=False)
            case AllAgentAction.unpause_on_resume:
                await self._set_unpause_on_resume(env, should_be_unpaused_on_resume=True)
            case AllAgentAction.remove_all_agent_venvs:
                await self._remove_executor_venvs(env)
            case _ as _never:
                assert_never(_never)

    async def _remove_executor_venvs(self, env: data.Environment) -> None:
        """
        Remove the venvs of the executors used by the given environment.
        """
        agent_client = self.get_agent_client(tid=env.id, endpoint=AGENT_SCHEDULER_ID, live_agent_only=False)
        if agent_client:
            self.add_background_task(agent_client.remove_executor_venvs())
        else:
            raise Conflict(f"No scheduler process is running for environment {env.id}")

    @handle(methods_v2.agent_action, env="tid")
    async def agent_action(self, env: data.Environment, name: str, action: AgentAction) -> None:
        if name == const.AGENT_SCHEDULER_ID:
            raise BadRequest(f"Particular action cannot be directed towards the Scheduler agent: {action.name}")

        if env.halted and action in {AgentAction.pause, AgentAction.unpause}:
            raise Forbidden("Can not pause or unpause agents when the environment has been halted.")
        if not env.halted and action in {AgentAction.keep_paused_on_resume, AgentAction.unpause_on_resume}:
            raise Forbidden("Cannot set on_resume state of agents when the environment is not halted.")
        match action:
            case AgentAction.pause:
                await self._pause_agent(env, name)
            case AgentAction.unpause:
                await self._unpause_agent(env, name)
            case AgentAction.keep_paused_on_resume:
                await self._set_unpause_on_resume(env, should_be_unpaused_on_resume=False, endpoint=name)
            case AgentAction.unpause_on_resume:
                await self._set_unpause_on_resume(env, should_be_unpaused_on_resume=True, endpoint=name)
            case _ as _never:
                assert_never(_never)

    async def _update_paused_status_agent(
        self,
        env: data.Environment,
        new_paused_status: bool,
        endpoint: Optional[str] = None,
        *,
        connection: Optional[asyncpg.connection.Connection] = None,
    ) -> None:
        """
        Helper method to pause / unpause a logical agent by pausing an active agent instance if it exists and notify the
        scheduler that something has changed.

        If no endpoint provided, pauses all logical agents. This does not include the scheduler itself.
        """
        # We need this lock otherwise, we would have transaction conflict in DB
        async with self.session_lock:
            await data.Agent.pause(env=env.id, endpoint=endpoint, paused=new_paused_status, connection=connection)
            key = (env.id, const.AGENT_SCHEDULER_ID)
            live_session = self.tid_endpoint_to_session.get(key)
            if live_session:
                await live_session.get_client().set_state(endpoint, enabled=not new_paused_status)

    async def _pause_agent(
        self, env: data.Environment, endpoint: Optional[str] = None, connection: Optional[asyncpg.connection.Connection] = None
    ) -> None:
        """
        Pause a logical agent by pausing an active agent instance if it exists.
        If no endpoint provided, pauses all logical agents. This does not include the scheduler itself.
        """
        await self._update_paused_status_agent(env=env, new_paused_status=True, endpoint=endpoint, connection=connection)

    async def _unpause_agent(
        self, env: data.Environment, endpoint: Optional[str] = None, connection: Optional[asyncpg.connection.Connection] = None
    ) -> None:
        """
        Unpause a logical agent by pausing an active agent instance if it exists.
        If no endpoint provided, pauses all logical agents. This does not include the scheduler itself.
        """
        await self._update_paused_status_agent(env=env, new_paused_status=False, endpoint=endpoint, connection=connection)

    async def _set_unpause_on_resume(
        self,
        env: data.Environment,
        should_be_unpaused_on_resume: bool,
        endpoint: Optional[str] = None,
        connection: Optional[asyncpg.connection.Connection] = None,
    ) -> None:
        """
        Set the unpause_on_resume field of an agent (or all agents in an environment when the endpoint is set to None)
        so that the agent is paused or unpaused after the environment is resumed
        """
        await data.Agent.set_unpause_on_resume(
            env=env.id, endpoint=endpoint, should_be_unpaused_on_resume=should_be_unpaused_on_resume, connection=connection
        )

    async def _process_session_listener_actions(self) -> None:
        """
        This is the consumer of the _session_listener_actions queue.
        """
        while not self.is_stopping():
            try:
                session_action = await self._session_listener_actions.get()
                await self._process_action(session_action)
            except asyncio.CancelledError:
                return
            except Exception:
                LOGGER.exception(
                    "An exception occurred while handling session action %s on session id %s.",
                    session_action.action_type.name,
                    session_action.session.id,
                    exc_info=True,
                )
            finally:
                try:
                    self._session_listener_actions.task_done()
                except Exception:
                    # Should never occur
                    pass

    async def _process_action(self, action: SessionAction) -> None:
        """
        Process a specific SessionAction.
        """
        action_type = action.action_type
        if action_type == SessionActionType.REGISTER_SESSION:
            await self._register_session(action.session, action.endpoint_names_snapshot, action.timestamp)
        elif action_type == SessionActionType.SEEN_SESSION:
            await self._seen_session(action.session, action.endpoint_names_snapshot)
        elif action_type == SessionActionType.EXPIRE_SESSION:
            await self._expire_session(action.session, action.endpoint_names_snapshot, action.timestamp)
        else:
            LOGGER.warning("Unknown SessionAction %s", action_type.name)

    # Notify from session listener
    async def new_session(self, session: protocol.Session, endpoint_names_snapshot: set[str]) -> None:
        """
        The _session_listener_actions queue ensures that all SessionActions are executed in the order of arrival.
        """
        session_action = SessionAction(
            action_type=SessionActionType.REGISTER_SESSION,
            session=session,
            endpoint_names_snapshot=endpoint_names_snapshot,
            timestamp=datetime.now().astimezone(),
        )
        await self._session_listener_actions.put(session_action)

    # Notify from session listener
    async def expire(self, session: protocol.Session, endpoint_names_snapshot: set[str]) -> None:
        """
        The _session_listener_actions queue ensures that all SessionActions are executed in the order of arrival.
        """
        session_action = SessionAction(
            action_type=SessionActionType.EXPIRE_SESSION,
            session=session,
            endpoint_names_snapshot=endpoint_names_snapshot,
            timestamp=datetime.now().astimezone(),
        )
        await self._session_listener_actions.put(session_action)

    # Notify from session listener
    async def seen(self, session: protocol.Session, endpoint_names_snapshot: set[str]) -> None:
        """
        The _session_listener_actions queue ensures that all SessionActions are executed in the order of arrival.
        """
        session_action = SessionAction(
            action_type=SessionActionType.SEEN_SESSION,
            session=session,
            endpoint_names_snapshot=endpoint_names_snapshot,
            timestamp=datetime.now().astimezone(),
        )
        await self._session_listener_actions.put(session_action)

    # Seen
    @tracing.instrument("AgentManager.seen_session", extract_args=True)
    async def _seen_session(self, session: protocol.Session, endpoint_names_snapshot: set[str]) -> None:
        endpoints_with_new_primary: list[tuple[str, Optional[uuid.UUID]]] = []
        async with self.session_lock:
            endpoints_in_agent_manager = self.endpoints_for_sid[session.id]
            endpoints_in_session = endpoint_names_snapshot
            endpoints_to_add = endpoints_in_session - endpoints_in_agent_manager
            LOGGER.debug("Adding endpoints %s to session %s on %s", endpoints_to_add, session.id, session.nodename)
            endpoints_to_remove = endpoints_in_agent_manager - endpoints_in_session
            LOGGER.debug("Removing endpoints %s from session %s on %s", endpoints_to_remove, session.id, session.nodename)

            endpoints_with_new_primary += await self._failover_endpoints(session, endpoints_to_remove)
            endpoints_with_new_primary += await self._ensure_primary_if_not_exists(session, endpoints_to_add)
            self.endpoints_for_sid[session.id] = endpoints_in_session

        self.add_background_task(
            self._log_session_seen_to_db(session, endpoints_to_add, endpoints_to_remove, endpoints_with_new_primary)
        )

    async def _log_session_seen_to_db(
        self,
        session: protocol.Session,
        endpoints_to_add: set[str],
        endpoints_to_remove: set[str],
        endpoints_with_new_primary: list[tuple[str, Optional[uuid.UUID]]],
    ) -> None:
        """
        Note: This method call is allowed to fail when the database connection is lost.
        """
        now = datetime.now().astimezone()
        async with data.AgentProcess.get_connection() as connection:
            async with connection.transaction():
                await data.AgentProcess.update_last_seen(session.id, now, connection)
                await data.AgentInstance.log_instance_creation(session.tid, session.id, endpoints_to_add, connection)
                await data.AgentInstance.log_instance_expiry(session.id, endpoints_to_remove, now, connection)
                await data.Agent.update_primary(session.tid, endpoints_with_new_primary, now, connection)

    # Session registration
    async def _register_session(self, session: protocol.Session, endpoint_names_snapshot: set[str], now: datetime) -> None:
        """
        This method registers a new session in memory and asynchronously updates the agent
        session log in the database. When the database connection is lost, the get_statuses()
        call fails and the new session will be refused.
        """
        LOGGER.debug("New session %s for agents %s on %s", session.id, endpoint_names_snapshot, session.nodename)
        async with self.session_lock:
            tid = session.tid
            sid = session.get_id()
            self.sessions[sid] = session
            self.endpoints_for_sid[sid] = endpoint_names_snapshot
            try:
                endpoints_with_new_primary = await self._ensure_primary_if_not_exists(session, endpoint_names_snapshot)
            except Exception as e:
                # Database connection failed
                del self.sessions[sid]
                del self.endpoints_for_sid[sid]
                self.add_background_task(session.expire(timeout=0))
                raise e

        self.add_background_task(
            self._log_session_creation_to_db(tid, session, endpoint_names_snapshot, endpoints_with_new_primary, now)
        )

    async def _log_session_creation_to_db(
        self,
        tid: uuid.UUID,
        session: protocol.Session,
        endpoint_names: set[str],
        endpoints_with_new_primary: Sequence[tuple[str, Optional[uuid.UUID]]],
        now: datetime,
    ) -> None:
        """
        Note: This method call is allowed to fail when the database connection is lost.
        """
        async with data.AgentProcess.get_connection() as connection:
            async with connection.transaction():
                await data.AgentProcess.seen(tid, session.nodename, session.id, now, connection)
                await data.AgentInstance.log_instance_creation(tid, session.id, endpoint_names, connection)
                await data.Agent.update_primary(tid, endpoints_with_new_primary, now, connection)

    # Session expiry
    async def _expire_session(self, session: protocol.Session, endpoint_names_snapshot: set[str], now: datetime) -> None:
        """
        This method expires the given session and update the in-memory session state.
        The in-database session log is updated asynchronously. These database updates
        are allowed to fail when the database connection is lost.
        """
        if not self.is_running() or self.is_stopping():
            return
        async with self.session_lock:
            tid = session.tid
            sid = session.get_id()
            if sid not in self.sessions:
                # The session is already expired
                return
            LOGGER.debug("expiring session %s", sid)
            del self.sessions[sid]
            del self.endpoints_for_sid[sid]
            endpoints_with_new_primary = await self._failover_endpoints(session, endpoint_names_snapshot)

        self.add_background_task(self._log_session_expiry_to_db(tid, endpoints_with_new_primary, session, now))

    async def _log_session_expiry_to_db(
        self,
        tid: uuid.UUID,
        endpoints_with_new_primary: Sequence[tuple[str, Optional[uuid.UUID]]],
        session: protocol.Session,
        now: datetime,
    ) -> None:
        """
        Note: This method call is allowed to fail when the database connection is lost.
        """
        async with data.AgentProcess.get_connection() as connection:
            async with connection.transaction():
                # Make sure to access the database tables in the order defined in docs string of inmanta/data/__init__.py
                # to prevent deadlock issues.
                await data.AgentProcess.expire_process(session.id, now, connection)
                await data.AgentInstance.log_instance_expiry(session.id, session.endpoint_names, now, connection)
                await data.Agent.update_primary(tid, endpoints_with_new_primary, now, connection)

    async def _expire_all_sessions_in_db(self) -> None:
        async with self.session_lock:
            LOGGER.debug("Cleaning server session DB")
            async with data.AgentProcess.get_connection() as connection:
                async with connection.transaction():
                    await data.AgentProcess.expire_all(now=datetime.now().astimezone(), connection=connection)
                    await data.AgentInstance.expire_all(now=datetime.now().astimezone(), connection=connection)
                    await data.Agent.mark_all_as_non_primary(connection=connection)

    async def _purge_agent_processes(self) -> None:
        agent_processes_to_keep = opt.agent_processes_to_keep.get()
        await data.AgentProcess.cleanup(nr_expired_records_to_keep=agent_processes_to_keep)

    # Util
    async def _use_new_active_session_for_agent(self, tid: uuid.UUID, endpoint_name: str) -> Optional[protocol.Session]:
        """
        This method searches for a new active session for the given agent. If a new active session if found,
        the in-memory state of the agentmanager is updated to use that new session. No logging is done in the
        database.

        :return The new active session in use or None if no new active session was found

        Note: Always call under session lock.
        """
        key = (tid, endpoint_name)
        new_active_session = self._get_session_to_failover_agent(tid, endpoint_name)
        if new_active_session:
            self.tid_endpoint_to_session[key] = new_active_session
            set_state_call = new_active_session.get_client().set_state(endpoint_name, enabled=True)
            self.add_background_task(set_state_call)
        elif key in self.tid_endpoint_to_session:
            del self.tid_endpoint_to_session[key]
        return new_active_session

    async def _ensure_primary_if_not_exists(
        self, session: protocol.Session, endpoints: set[str]
    ) -> Sequence[tuple[str, uuid.UUID]]:
        """
        Make this session the primary session for the endpoints of this session if no primary exists and the agent is not
        paused.

        :return: The endpoints that got a new primary.

        Note: Always call under session lock.
        Note: This call will fail when the database connection is lost.
        """
        agent_statuses = await data.Agent.get_statuses(session.tid, endpoints)

        result = []
        for endpoint in endpoints:
            key = (session.tid, endpoint)
            if key not in self.tid_endpoint_to_session and agent_statuses[endpoint] != AgentStatus.paused:
                LOGGER.debug("set session %s as primary for agent %s in env %s", session.id, endpoint, session.tid)
                self.tid_endpoint_to_session[key] = session
                self.add_background_task(session.get_client().set_state(endpoint, enabled=True))
                result.append((endpoint, session.id))
        return result

    async def _failover_endpoints(
        self, session: protocol.Session, endpoints: set[str]
    ) -> Sequence[tuple[str, Optional[uuid.UUID]]]:
        """
        If the given session is the primary for a given endpoint, failover to a new session.

        :return: The endpoints that got a new primary.

        Note: Always call under session lock.
        """
        agent_statuses = await data.Agent.get_statuses(session.tid, endpoints)
        result = []
        for endpoint_name in endpoints:
            key = (session.tid, endpoint_name)
            if key in self.tid_endpoint_to_session and self.tid_endpoint_to_session[key].id == session.id:
                if agent_statuses[endpoint_name] != AgentStatus.paused:
                    new_active_session = await self._use_new_active_session_for_agent(session.tid, endpoint_name)
                    if new_active_session:
                        result.append((endpoint_name, new_active_session.id))
                    else:
                        result.append((endpoint_name, None))
                else:
                    # This should never occur. An agent cannot have an active session while its paused,
                    # given the fact that this method executes under session_lock
                    LOGGER.warning("Paused agent %s has an active session (sid=%s)", endpoint_name, session.id)
                    del self.tid_endpoint_to_session[key]
                    result.append((endpoint_name, None))
        return result

    def is_primary(self, env: data.Environment, sid: uuid.UUID, agent: str) -> bool:
        prim = self.tid_endpoint_to_session.get((env.id, agent), None)
        if not prim:
            return False
        return prim.get_id() == sid

    def get_session_for(self, tid: uuid.UUID, endpoint: str) -> Optional[protocol.Session]:
        """
        Return a session that matches the given environment and endpoint.
        This method also returns session to paused or non-live agents.
        """
        key = (tid, endpoint)
        session = self.tid_endpoint_to_session.get(key)
        if session:
            # Agent has live session
            return session
        else:
            # Maybe session exists for a paused agent
            for session in self.sessions.values():
                if endpoint in session.endpoint_names and session.tid == tid:
                    return session
            # Agent is down
            return None

    def _get_session_to_failover_agent(self, tid: uuid.UUID, endpoint: str) -> Optional[protocol.Session]:
        current_active_session = self.tid_endpoint_to_session[(tid, endpoint)]
        for session in self.sessions.values():
            if endpoint in session.endpoint_names and session.tid == tid:
                if not current_active_session or session.id != current_active_session.id:
                    return session
        return None

    def get_agent_client(self, tid: uuid.UUID, endpoint: str, live_agent_only: bool = True) -> Optional[ReturnClient]:
        if isinstance(tid, str):
            tid = uuid.UUID(tid)
        key = (tid, endpoint)
        session = self.tid_endpoint_to_session.get(key)
        if session:
            return session.get_client()
        elif not live_agent_only:
            session = self.get_session_for(tid, endpoint)
            if session:
                return session.get_client()
            else:
                return None
        else:
            return None

    async def expire_sessions_for_agents(self, env_id: uuid.UUID, endpoints: Set[str]) -> None:
        """
        Expire all sessions for any of the requested agent endpoints.
        """
        async with self.session_lock:
            sessions_to_expire: Iterator[protocol.Session] = (
                session for session in self.sessions.values() if endpoints & session.endpoint_names and session.tid == env_id
            )
            await asyncio.gather(*(s.expire_and_abort(timeout=0) for s in sessions_to_expire))

    async def are_agents_active(self, tid: uuid.UUID, endpoints: Iterable[str]) -> bool:
        """
        Return true iff all the given agents are in the up or the paused state.
        """
        return all(active for (_, active) in await self.get_agent_active_status(tid, endpoints))

    async def get_agent_active_status(self, tid: uuid.UUID, endpoints: Iterable[str]) -> list[tuple[str, bool]]:
        """
        Return a list of tuples where the first element of the tuple contains the name of an endpoint
        and the second a boolean indicating where there is an active (up or paused) agent for that endpoint.
        """
        all_sids_for_env = [sid for (sid, session) in self.sessions.items() if session.tid == tid]
        all_active_endpoints_for_env = {ep for sid in all_sids_for_env for ep in self.endpoints_for_sid[sid]}
        return [(ep, ep in all_active_endpoints_for_env) for ep in endpoints]

    async def expire_all_sessions_for_environment(self, env_id: uuid.UUID) -> None:
        async with self.session_lock:
            await asyncio.gather(*[s.expire_and_abort(timeout=0) for s in self.sessions.values() if s.tid == env_id])

    async def expire_all_sessions(self) -> None:
        async with self.session_lock:
            await asyncio.gather(*[s.expire_and_abort(timeout=0) for s in self.sessions.values()])

    # Agent Management
    @tracing.instrument("AgentManager.ensure_agent_registered")
    async def ensure_agent_registered(
        self, env: data.Environment, nodename: str, *, connection: Optional[asyncpg.connection.Connection] = None
    ) -> data.Agent:
        """
        Make sure that an agent has been created in the database
        """
        async with self.session_lock:
            agent = await data.Agent.get(env.id, nodename, connection=connection)
            if agent:
                return agent
            else:
                return await self._create_default_agent(env, nodename, connection=connection)

    async def _create_default_agent(
        self, env: data.Environment, nodename: str, *, connection: Optional[asyncpg.connection.Connection] = None
    ) -> data.Agent:
        """
        This method creates a new agent (agent in the model) in the database.
        If an active agent instance exists for the given agent, it is marked as the
        primary instance for that agent in the database.

        Note: This method must be called under session lock
        """
        saved = data.Agent(environment=env.id, name=nodename, paused=False)
        await saved.insert(connection=connection)

        key = (env.id, nodename)
        session = self.tid_endpoint_to_session.get(key)
        if session:
            await data.Agent.update_primary(
                env.id, [(nodename, session.id)], datetime.now().astimezone(), connection=connection
            )

        return saved

    # External APIS
    @handle(methods.get_agent_process, agent_sid="id")
    async def get_agent_process(self, agent_sid: uuid.UUID) -> Apireturn:
        return await self.get_agent_process_report(agent_sid)

    @handle(methods.list_agent_processes)
    async def list_agent_processes(
        self,
        environment: Optional[UUID],
        expired: bool,
        start: Optional[UUID] = None,
        end: Optional[UUID] = None,
        limit: Optional[int] = None,
    ) -> Apireturn:
        """List all agent processes whose sid is after start and before end

        :param environment: Optional, the environment the agent should come from.
        :param expired: If True, expired agents will also be shown, they are hidden otherwise.
        :param start: The sid all of the selected agent process should be greater than this, defaults to None
        :param end: The sid all of the selected agent process should be smaller than this, defaults to None
        :param limit: Whether to limit the number of returned entries, defaults to None
        :raises BadRequest: Limit, start and end can not be set together
        :raises NotFound: The given environment id does not exist!
        :raises BadRequest: Limit parameter can not exceed 1000
        """
        query: dict[str, Any] = {}
        if environment is not None:
            query["environment"] = environment
            env = await data.Environment.get_by_id(environment)
            if env is None:
                return 404, {"message": "The given environment id does not exist!"}
        if not expired:
            query["expired"] = None

        if limit is None:
            limit = APILIMIT
        elif limit > APILIMIT:
            raise BadRequest(f"Limit parameter can not exceed {APILIMIT}, got {limit}.")

        aps = await data.AgentProcess.get_list_paged(
            page_by_column="sid",
            limit=limit,
            start=start,
            end=end,
            no_obj=False,
            connection=None,
            **query,
        )

        processes = []
        for p in aps:
            agent_dict = p.to_dict()
            ais = await data.AgentInstance.get_list(process=p.sid)
            oais = []
            for ai in ais:
                a = ai.to_dict()
                oais.append(a)
            agent_dict["endpoints"] = oais
            processes.append(agent_dict)

        return 200, {"processes": processes}

    @handle(methods.list_agents, env="tid")
    async def list_agents(
        self,
        env: data.Environment,
        start: Optional[str] = None,
        end: Optional[str] = None,
        limit: Optional[int] = None,
    ) -> Apireturn:
        """List all agents whose name is after start and before end

        :param env: The environment the agents should come from
        :param start: The name all of the selected agent should be greater than this, defaults to None
        :param end: The name all of the selected agent should be smaller than this, defaults to None
        :param limit: Whether to limit the number of returned entries, defaults to None
        :raises BadRequest: Limit, start and end can not be set together
        :raises BadRequest: Limit parameter can not exceed 1000
        """
        if env is None:
            return 404, {"message": "The given environment id does not exist!"}
        new_agent_endpoint = await self.get_agents(env, limit, start, end, start, end)

        def mangle_format(agent: model.Agent) -> dict[str, object]:
            native = agent.model_dump()
            native["primary"] = ""
            native["state"] = agent.status
            if native["last_failover"] is None:
                native["last_failover"] = ""
            return native

        return 200, {
            "agents": [mangle_format(a) for a in new_agent_endpoint._response],
            "servertime": datetime.now().astimezone(),
        }

    @handle(methods.get_state, env="tid")
    async def get_state(self, env: data.Environment, sid: uuid.UUID, agent: str) -> Apireturn:
        tid: UUID = env.id
        if isinstance(tid, str):
            tid = uuid.UUID(tid)
        key = (tid, agent)
        session = self.tid_endpoint_to_session.get(key, None)
        if session is not None and session.id == sid:
            return 200, {"enabled": True}
        return 200, {"enabled": False}

    async def get_agent_process_report(self, agent_sid: uuid.UUID) -> ReturnTupple:
        ap = await data.AgentProcess.get_one(sid=agent_sid)
        if ap is None:
            return 404, {"message": "The given AgentProcess id does not exist!"}
        sid = ap.sid
        session_for_ap = self.sessions.get(sid, None)
        if session_for_ap is None:
            return 404, {"message": "The given AgentProcess is not live!"}
        client = session_for_ap.get_client()
        result = await client.get_status()
        return result.code, result.get_result()

    async def request_parameter(self, env_id: uuid.UUID, resource_id: ResourceIdStr) -> Apireturn:
        """
        Request the value of a parameter from an agent
        """
        if resource_id is not None and resource_id != "":
            env = await data.Environment.get_by_id(env_id)

            if env is None:
                raise NotFound(f"Environment with {env_id} does not exist.")

            # get a resource version
            res = await data.Resource.get_latest_version(env_id, resource_id)

            if res is None:
                return 404, {"message": "The resource has no recent version."}

            if res.status in UNDEPLOYABLE_NAMES:
                LOGGER.debug(
                    "Ignore fact request for %s, resource is in an undeployable state.",
                    resource_id,
                )
                return 503, {"message": "The resource is in an undeployable state."}

            # only request facts of a resource every _fact_resource_block time
            now = time.time()
            if (
                resource_id not in self._fact_resource_block_set
                or (self._fact_resource_block_set[resource_id] + self._fact_resource_block) < now
            ):
                await self._autostarted_agent_manager._ensure_scheduler(env_id)
                agent = const.AGENT_SCHEDULER_ID
                client = self.get_agent_client(env_id, agent)
                if client is not None:
                    await client.get_parameter(env_id, agent, res.to_dict())

                self._fact_resource_block_set[resource_id] = now

            else:
                LOGGER.debug(
                    "Ignore fact request for %s, last request was sent %d seconds ago.",
                    resource_id,
                    now - self._fact_resource_block_set[resource_id],
                )

            return 503, {"message": "Agents queried for resource parameter."}
        else:
            return 404, {"message": "resource_id parameter is required."}

    @handle(methods_v2.get_agents, env="tid")
    async def get_agents(
        self,
        env: data.Environment,
        limit: Optional[int] = None,
        start: Optional[Union[datetime, bool, str]] = None,
        end: Optional[Union[datetime, bool, str]] = None,
        first_id: Optional[str] = None,
        last_id: Optional[str] = None,
        filter: Optional[dict[str, list[str]]] = None,
        sort: str = "name.asc",
    ) -> ReturnValue[Sequence[model.Agent]]:
        try:
            handler = AgentView(
                environment=env,
                limit=limit,
                sort=sort,
                first_id=first_id,
                last_id=last_id,
                start=start,
                end=end,
                filter=filter,
            )
            out = await handler.execute()
            return out
        except (InvalidFilter, InvalidSort, data.InvalidQueryParameter, data.InvalidFieldNameException) as e:
            raise BadRequest(e.message) from e

    @handle(methods_v2.get_agent_process_details, env="tid")
    async def get_agent_process_details(self, env: data.Environment, id: uuid.UUID, report: bool = False) -> model.AgentProcess:
        agent_process = await data.AgentProcess.get_one(environment=env.id, sid=id)
        if not agent_process:
            raise NotFound(f"Agent process with id {id} not found")
        dto = agent_process.to_dto()
        if report:
            report_status, report_result = await self.get_agent_process_report(id)
            if report_status == 200:
                dto.state = report_result
        return dto


@dataclass
class ProcessDetails:
    """
    A dataclass that holds the details of a process.
    """

    process: subprocess.Process
    path_stdout: str
    path_stderr: str

    @property
    def pid(self) -> int:
        """
        Returns the process id of the process.
        """
        return self.process.pid

    def is_running(self) -> bool:
        """
        Return True iff the process is running.
        """
        return self.process.returncode is None


class AutostartedAgentManager(ServerSlice, inmanta.server.services.environmentlistener.EnvironmentListener):
    """
    An instance of this class manages scheduler processes.
    """

    environment_service: "environmentservice.EnvironmentService"

    def __init__(self) -> None:
        super().__init__(SLICE_AUTOSTARTED_AGENT_MANAGER)
        self._agent_procs: dict[UUID, ProcessDetails] = {}  # env uuid -> ProcessDetails
        self.agent_lock = asyncio.Lock()  # Prevent concurrent updates on _agent_procs

    async def get_status(self) -> Mapping[str, ArgumentTypes]:
        return {"processes": len(self._agent_procs)}

    async def prestart(self, server: protocol.Server) -> None:
        await ServerSlice.prestart(self, server)
        preserver = server.get_slice(SLICE_SERVER)
        assert isinstance(preserver, Server)
        self._server: Server = preserver
        self._server_storage: dict[str, str] = self._server._server_storage

        agent_manager = server.get_slice(SLICE_AGENT_MANAGER)
        assert isinstance(agent_manager, AgentManager)
        self._agent_manager = agent_manager

        self.environment_service = cast(environmentservice.EnvironmentService, server.get_slice(SLICE_ENVIRONMENT))
        self.environment_service.register_listener(self, inmanta.server.services.environmentlistener.EnvironmentAction.created)

    async def start(self) -> None:
        await super().start()
        self.add_background_task(self._start_agents())

    async def prestop(self) -> None:
        await super().prestop()
        await self._terminate_agents()

    async def stop(self) -> None:
        await super().stop()

    def get_dependencies(self) -> list[str]:
        return [SLICE_SERVER, SLICE_DATABASE, SLICE_AGENT_MANAGER]

    def get_depended_by(self) -> list[str]:
        return [SLICE_TRANSPORT]

    async def _start_agents(self) -> None:
        """
        Ensure that a scheduler is started for each environment having AUTOSTART_ON_START is true. This method
        is called on server start.
        """
        environments = await data.Environment.get_list()
        for env in environments:
            autostart = await env.get(data.AUTOSTART_ON_START)
            if not autostart:
                continue
            await self._ensure_scheduler(env.id)

    async def restart_agents(self, env: data.Environment) -> None:
        LOGGER.debug("Restarting Scheduler in environment %s", env.id)
        await self._ensure_scheduler(env.id)

    async def stop_agents(
        self,
        env: data.Environment,
        *,
        delete_venv: bool = False,
    ) -> None:
        """
        Stop all agents for this environment and close sessions
        """
        async with self.agent_lock:
            LOGGER.debug("Stopping scheduler for env %s", env.id)
            if env.id in self._agent_procs:
                proc_details = self._agent_procs[env.id]
                self._stop_process(proc_details)
                await self._wait_for_proc_bounded([proc_details])
                del self._agent_procs[env.id]
            if delete_venv:
                self._remove_venv_for_agent_in_env(env.id)

            LOGGER.debug("Expiring all sessions for %s", env.id)
            await self._agent_manager.expire_all_sessions_for_environment(env.id)

    async def _stop_scheduler(
        self,
        env: data.Environment,
        *,
        connection: Optional[asyncpg.connection.Connection] = None,
    ) -> None:
        """
        Stop the scheduler for this environment and expire all its sessions.

        Must be called under the agent lock
        """
        LOGGER.debug("Stopping scheduler for environment %s", env.id)
        if env.id in self._agent_procs:
            proc_details = self._agent_procs[env.id]
            self._stop_process(proc_details)
            await self._wait_for_proc_bounded([proc_details])
            del self._agent_procs[env.id]

        LOGGER.debug("Expiring session for scheduler in environment %s", env.id)
        await self._agent_manager.expire_sessions_for_agents(env.id, endpoints={AGENT_SCHEDULER_ID})

    def _get_state_dir_for_agent_in_env(self, env_id: uuid.UUID) -> str:
        """
        Return the state dir to be used by the auto-started agent in the given environment.
        """
        state_dir: str = inmanta.config.state_dir.get()
        return os.path.join(state_dir, "server", str(env_id))

    def _remove_venv_for_agent_in_env(self, env_id: uuid.UUID) -> None:
        """
        Remove the venv for the auto-started agent in the given environment.
        """
        agent_state_dir: str = self._get_state_dir_for_agent_in_env(env_id)
        venv_dir: str = os.path.join(agent_state_dir, "agent", "env")
        try:
            shutil.rmtree(venv_dir)
        except FileNotFoundError:
            pass

    def _stop_process(self, process_details: ProcessDetails) -> None:
        try:
            process_details.process.terminate()
        except ProcessLookupError:
            # Process was already terminated
            pass

    async def _terminate_agents(self) -> None:
        async with self.agent_lock:
            LOGGER.debug("Stopping all schedulers")
            for proc_details in self._agent_procs.values():
                self._stop_process(proc_details)
            await self._wait_for_proc_bounded(self._agent_procs.values())
            LOGGER.debug("Expiring all sessions")
            await self._agent_manager.expire_all_sessions()

    # Start/Restart scheduler
    async def _ensure_scheduler(
        self,
        env: uuid.UUID,
        *,
        restart: bool = False,
        connection: Optional[asyncpg.connection.Connection] = None,
    ) -> bool:
        """
        Ensure that the scheduler for the given environment, that should be autostarted, are started.

        :param env: The environment to start the scheduler for.
        :param restart: Restart the scheduler even if it's running.
        :param connection: The database connection to use. Must not be in a transaction context.

        :return: True iff a new scheduler process was started.
        """
        if no_start_scheduler:
            return False
        if self._stopping:
            raise ShutdownInProgress()

        if connection is not None and connection.is_in_transaction():
            # Should not be called in a transaction context because it has (immediate) side effects outside of the database
            # that are tied to the database state. Several inconsistency issues could occur if this runs in a transaction
            # context:
            #   - side effects based on oncommitted reads (may even need to be rolled back)
            #   - race condition with similar side effect flows due to stale reads (e.g. other flow pauses agent and kills
            #       process, this one brings it back because it reads the agent as unpaused)
            raise Exception("_ensure_scheduler should not be called in a transaction context")

        autostart_scheduler = {const.AGENT_SCHEDULER_ID}
        async with data.Agent.get_connection(connection) as connection:
            async with self.agent_lock:
                # silently ignore requests if this environment is halted
                refreshed_env: Optional[data.Environment] = await data.Environment.get_by_id(env, connection=connection)
                if refreshed_env is None:
                    raise inmanta.exceptions.EnvironmentNotFound(f"Can't ensure scheduler: environment {env} does not exist")
                if refreshed_env.halted:
                    return False

                are_active = await self._agent_manager.are_agents_active(env, autostart_scheduler)
                if not restart and are_active:
                    # do not start a new agent process if the agents are already active, regardless of whether their session
                    # is with an autostarted process or not.
                    return False

                start_new_process: bool
                if env not in self._agent_procs or not self._agent_procs[env].is_running():
                    # Start new process if none is currently running for this environment.
                    LOGGER.info("%s matches agents managed by server, ensuring they are started.", autostart_scheduler)
                    start_new_process = True
                elif restart:
                    LOGGER.info(
                        "%s matches agents managed by server, forcing restart: stopping process with PID %s.",
                        autostart_scheduler,
                        self._agent_procs[env].process,
                    )
                    await self._stop_scheduler(refreshed_env, connection=connection)
                    start_new_process = True
                else:
                    start_new_process = False

                if start_new_process:
                    self._agent_procs[env] = await self.__do_start_agent(refreshed_env, connection=connection)

                # Wait for all agents to start
                try:
                    await self._wait_for_agents(refreshed_env, autostart_scheduler, connection=connection)
                except asyncio.TimeoutError:
                    LOGGER.warning("Not all agent instances started successfully")
                return start_new_process

    async def __do_start_agent(
        self, env: data.Environment, *, connection: Optional[asyncpg.connection.Connection] = None
    ) -> ProcessDetails:
        """
        Start an autostarted agent process for the given environment. Should only be called if none is running yet.

        :return: A ProcessDetails object consisting of the agent process, the stdout log file and the stderr log file.
        """
        assert not assert_no_start_scheduler

        config: str = await self._make_agent_config(env, connection=connection)

        root_dir: str = self._server_storage["server"]
        config_dir = os.path.join(root_dir, str(env.id))

        if not os.path.exists(config_dir):
            os.mkdir(config_dir)

        config_path = os.path.join(config_dir, "scheduler.cfg")
        with open(config_path, "w+", encoding="utf-8") as fd:
            fd.write(config)

        out: str = os.path.join(self._server_storage["logs"], "agent-%s.out" % env.id)
        err: str = os.path.join(self._server_storage["logs"], "agent-%s.err" % env.id)

        proc: subprocess.Process = await self._fork_inmanta(
            [
                "--log-file-level",
                "DEBUG",
                "--config",
                config_path,
                *(["--config-dir", Config._config_dir] if Config._config_dir is not None else []),
                "scheduler",
            ],
            out,
            err,
        )

        LOGGER.debug("Started new agent with PID %s", proc.pid)
        return ProcessDetails(process=proc, path_stdout=out, path_stderr=err)

    async def _make_agent_config(
        self,
        env: data.Environment,
        *,
        connection: Optional[asyncpg.connection.Connection],
    ) -> str:
        """
        Generate the config file for the process that hosts the autostarted agents

        :param env: The environment for which to autostart agents
        :return: A string that contains the config file content.
        """
        environment_id = str(env.id)
        port: int = opt.server_bind_port.get()

        privatestatedir: str = self._get_state_dir_for_agent_in_env(env.id)

        agent_deploy_interval: str = cast(str, await env.get(data.AUTOSTART_AGENT_DEPLOY_INTERVAL, connection=connection))

        agent_repair_interval: str = cast(str, await env.get(data.AUTOSTART_AGENT_REPAIR_INTERVAL, connection=connection))

        # generate config file
        config = f"""[config]
state-dir=%(statedir)s
log-dir={global_config.log_dir.get()}

environment=%(env_id)s

agent-deploy-interval=%(agent_deploy_interval)s
agent-repair-interval=%(agent_repair_interval)s

[agent]
executor-cap={agent_cfg.agent_executor_cap.get()}
executor-retention-time={agent_cfg.agent_executor_retention_time.get()}

[agent_rest_transport]
port=%(port)s
host={opt.internal_server_address.get()}
""" % {
            "env_id": environment_id,
            "port": port,
            "statedir": privatestatedir,
            "agent_deploy_interval": agent_deploy_interval,
            "agent_repair_interval": agent_repair_interval,
        }

        if server_config.server_enable_auth.get():
            token = encode_token(["agent"], environment_id)
            config += """
token=%s
    """ % (
                token
            )

        ssl_cert: Optional[str] = server_config.server_ssl_key.get()
        ssl_ca: Optional[str] = server_config.server_ssl_ca_cert.get()

        if ssl_ca is not None and ssl_cert is not None:
            # override CA
            config += """
ssl=True
ssl_ca_cert_file=%s
    """ % (
                ssl_ca
            )
        elif ssl_cert is not None:
            # system CA
            config += """
ssl=True
    """
        config += f"""
[database]
wait_time={opt.db_wait_time.get()}
host={opt.db_host.get()}
port={opt.db_port.get()}
name={opt.db_name.get()}
username={opt.db_username.get()}
password={opt.db_password.get()}

[scheduler]
db-connection-pool-min-size={agent_cfg.scheduler_db_connection_pool_min_size.get()}
db-connection-pool-max-size={agent_cfg.scheduler_db_connection_pool_max_size.get()}
db-connection-timeout={agent_cfg.scheduler_db_connection_timeout.get()}

[influxdb]
host = {opt.influxdb_host.get()}
port = {opt.influxdb_port.get()}
name = {opt.influxdb_name.get()}
username = {opt.influxdb_username.get()}
password = {opt.influxdb_password.get()}
interval = {opt.influxdb_interval.get()}
tags = {config_map_to_str(opt.influxdb_tags.get())}
"""

        if scheduler_log_config.get():
            config += f"""

[logging]
scheduler = {os.path.abspath(scheduler_log_config.get())}
"""

        return config

    async def _fork_inmanta(
        self, args: list[str], outfile: Optional[str], errfile: Optional[str], cwd: Optional[str] = None
    ) -> subprocess.Process:
        """
        Fork an inmanta process from the same code base as the current code
        """
        full_args = ["-m", "inmanta.app", *args]
        # handles can be closed, owned by child process,...
        outhandle = None
        errhandle = None
        try:
            if outfile is not None:
                outhandle = open(outfile, "wb+")
            if errfile is not None:
                errhandle = open(errfile, "wb+")

            env = os.environ.copy()
            env.update(tracing.get_context())
            return await asyncio.create_subprocess_exec(
                sys.executable, *full_args, cwd=cwd, env=env, stdout=outhandle, stderr=errhandle
            )
        finally:
            if outhandle is not None:
                outhandle.close()
            if errhandle is not None:
                errhandle.close()

    async def _wait_for_agents(
        self,
        env: data.Environment,
        agents: Set[str],
        *,
        connection: Optional[asyncpg.connection.Connection] = None,
    ) -> None:
        """
        Wait until all requested autostarted agent instances are active, e.g. after starting a new agent process.

        Must be called under the agent lock.

        :param env: The environment for which to wait for agents.
        :param agents: Autostarted agent endpoints to wait for.

        :raises TimeoutError: When not all agent instances are active and no new agent instance became active in the last
            5 seconds.
        """
        agent_statuses: dict[str, Optional[AgentStatus]] = await data.Agent.get_statuses(env.id, agents, connection=connection)
        # Only wait for agents that are not paused
        expected_agents_in_up_state: Set[str] = {
            agent_name
            for agent_name, status in agent_statuses.items()
            if status is not None and status is not AgentStatus.paused
        }

        assert env.id in self._agent_procs
        proc_details = self._agent_procs[env.id]

        actual_agents_in_up_state: set[str] = set()
        started = int(time.time())
        last_new_agent_seen = started
        last_log = started

        log_files = [
            inmanta_logging.LoggingConfigBuilder.get_log_file_for_scheduler(str(env.id), global_config.log_dir.get()),
            proc_details.path_stdout,
            proc_details.path_stderr,
        ]

        while len(expected_agents_in_up_state) != len(actual_agents_in_up_state):
            await asyncio.sleep(0.1)
            now = int(time.time())
            if now - last_new_agent_seen > AUTO_STARTED_AGENT_WAIT:
                LOGGER.warning(
                    "Timeout: agent with PID %s took too long to start: still waiting for agent instances %s."
                    " See log files %s for more information.",
                    proc_details.pid,
                    ",".join(sorted(expected_agents_in_up_state - actual_agents_in_up_state)),
                    ", ".join(log_files),
                )

                raise asyncio.TimeoutError()
            if now - last_log > AUTO_STARTED_AGENT_WAIT_LOG_INTERVAL:
                last_log = now
                LOGGER.debug(
                    "Waiting for agent with PID %s, waited %d seconds, %d/%d instances up",
                    proc_details.pid,
                    now - started,
                    len(actual_agents_in_up_state),
                    len(expected_agents_in_up_state),
                )
            new_actual_agents_in_up_state = {
                agent_name
                for agent_name in expected_agents_in_up_state
                if (
                    (session := self._agent_manager.tid_endpoint_to_session.get((env.id, agent_name), None)) is not None
                    # make sure to check for expiry because sessions are unregistered from the agent manager asynchronously
                    and not session.expired
                )
            }
            if len(new_actual_agents_in_up_state) > len(actual_agents_in_up_state):
                # Reset timeout timer because a new instance became active
                last_new_agent_seen = now
            actual_agents_in_up_state = new_actual_agents_in_up_state

        LOGGER.debug(
            "Agent process with PID %s is up for agent instances %s",
            proc_details.pid,
            ",".join(sorted(expected_agents_in_up_state)),
        )

    async def _wait_for_proc_bounded(
        self, proc_details: Iterable[ProcessDetails], timeout: float = const.SHUTDOWN_GRACE_HARD
    ) -> None:
        try:
            unfinished_processes = [p.process for p in proc_details if p.is_running()]
            await asyncio.wait_for(asyncio.gather(*[asyncio.shield(proc.wait()) for proc in unfinished_processes]), timeout)
        except asyncio.TimeoutError:
            LOGGER.warning("Agent processes did not close in time (%s)", [p.process for p in proc_details])

    async def environment_action_created(self, env: model.Environment) -> None:
        """
        Will be called when a new environment is created to create a scheduler agent

        :param env: The new environment
        """
        env_db = await data.Environment.get_by_id(env.id)
        assert env_db
        # We need to make sure that the AGENT_SCHEDULER is registered to be up and running
        await self._agent_manager.ensure_agent_registered(env_db, const.AGENT_SCHEDULER_ID)
        if not (assert_no_start_scheduler or no_start_scheduler):
            await self._ensure_scheduler(env.id)

    async def notify_agent_deploy_timer_update(self, env: data.Environment) -> None:
        agent_client = self._agent_manager.get_agent_client(tid=env.id, endpoint=AGENT_SCHEDULER_ID, live_agent_only=False)
        if agent_client:
            self.add_background_task(agent_client.notify_timer_update(env.id))


# For testing only
# Assert the scheduler will not be started (i.e. agent mock setup is correct)
assert_no_start_scheduler = False
# Ensure the scheduler is not started
no_start_scheduler = False



"""
Copyright 2019 Inmanta

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Contact: code@inmanta.com
"""

import asyncio
import json
import os
import pathlib
import time
from functools import partial

import jwt
import pytest
import tornado
from tornado import web

import requests
from inmanta import config, const
from inmanta.protocol import auth
from inmanta.server.protocol import Server


def test_jwt_create(inmanta_config):
    """
    Test creating, signing and verifying JWT with HS256 from the configuration
    """
    jot = auth.encode_token(["api"])
    payload, _ = auth.decode_token(jot)

    assert "api" in payload["urn:inmanta:ct"]

    # test creating an idempotent token
    jot1 = auth.encode_token(["agent"], idempotent=True)
    jot3 = auth.encode_token(["agent"])
    time.sleep(1)
    jot2 = auth.encode_token(["agent"], idempotent=True)
    jot4 = auth.encode_token(["agent"])
    assert jot1 == jot2
    assert jot3 != jot4
    assert jot1 != jot3
    assert jot2 != jot3


class PKHandler(web.RequestHandler):
    def get(self):
        public_key = {
            "keys": [
                {
                    "kid": "NMROYvAmygR5zzmtYXgmCbJd8cXHcvAtx_o-JxJNPoQ",
                    "kty": "RSA",
                    "alg": "RS256",
                    "use": "sig",
                    "n": "oA7zi9u230-e3atV4rW9oI6z3ea_rViKlBTqq4_v9E-PK47yPpUvnHH9eJrKBXcuX-cVO2pSmDQ65nAzEaobjBU8XPtm3sceY1GsC"
                    "cP4Uo7gEbqLxsaqN1WUt1tnbV10wEbZzHmtAW_J_J5wlIB696ceEdwxNyj3Zscq15QIMsahDmV54fvwussuPgNhd0t4ng9BzaW-kG"
                    "2-Z80blyxN3fUbXX1JRMOiX4a7W_UXN5Q9B4kE9vlqAm30FnhYvLLqKh-DFvxq49dbYTWR-pJSFkjRMD6u1MUzKQEmOLYTnDX42zA"
                    "rTVhbvQl7OnW-OXtwx9zVqiQIIqt5IQgZQ7PfwQ",
                    "e": "AQAB",
                }
            ]
        }

        self.write(json.dumps(public_key))


@pytest.fixture(scope="function")
async def jwks(unused_tcp_port):
    http_app = web.Application([(r"/auth/realms/inmanta/protocol/openid-connect/certs", PKHandler)])
    server = tornado.httpserver.HTTPServer(http_app)
    server.bind(unused_tcp_port)
    server.start()
    yield server

    server.stop()
    await server.close_all_connections()


async def test_validate_rs256(jwks, tmp_path):
    """
    Test that inmanta can download a rs256 public key
    """
    port = str(list(jwks._sockets.values())[0].getsockname()[1])
    config_file = os.path.join(tmp_path, "auth.cfg")
    with open(config_file, "w+", encoding="utf-8") as fd:
        fd.write(
            """
[auth_jwt_default]
algorithm=HS256
sign=true
client_types=agent,compiler
key=eciwliGyqECVmXtIkNpfVrtBLutZiITZKSKYhogeHMM
expire=0
issuer=https://localhost:8888/
audience=https://localhost:8888/

[auth_jwt_keycloak]
algorithm=RS256
sign=false
client_types=api
issuer=https://localhost:{0}/auth/realms/inmanta
audience=sodev
jwks_uri=http://localhost:{0}/auth/realms/inmanta/protocol/openid-connect/certs
validate_cert=false
""".format(
                port
            )
        )

    config.Config.load_config(config_file)

    cfg_list = await asyncio.get_event_loop().run_in_executor(None, auth.AuthJWTConfig.list)
    assert len(cfg_list) == 2


class SlowHandler(web.RequestHandler):
    async def get(self):
        await asyncio.sleep(5)
        self.write(json.dumps({"keys": "not only slow, but also invalid"}))


@pytest.fixture(scope="function")
async def slow_jwks(unused_tcp_port):
    http_app = web.Application([(r"/auth/realms/inmanta/protocol/openid-connect/certs", SlowHandler)])
    server = tornado.httpserver.HTTPServer(http_app)
    server.bind(unused_tcp_port)
    server.start()
    yield server

    server.stop()
    await server.close_all_connections()


async def test_rs256_invalid_config_timeout(tmp_path, slow_jwks):
    """
    Test that an error is raised when the timeout to download the rs256 public key is exceeded
    """
    port = str(list(slow_jwks._sockets.values())[0].getsockname()[1])
    config_file = os.path.join(tmp_path, "auth.cfg")
    with open(config_file, "w+", encoding="utf-8") as fd:
        fd.write(
            """
[auth_jwt_keycloak]
algorithm=RS256
sign=false
client_types=api
issuer=https://localhost:{0}/auth/realms/inmanta
audience=sodev
jwks_uri=http://localhost:{0}/auth/realms/inmanta/protocol/openid-connect/certs
jwks_request_timeout=0.1
validate_cert=false
""".format(
                port
            )
        )

    config.Config.load_config(config_file)
    with pytest.raises(ValueError):
        await asyncio.get_event_loop().run_in_executor(None, partial(auth.AuthJWTConfig.get, "auth_jwt_keycloak"))


def test_custom_claim_matching(tmp_path) -> None:
    """Validate parsing claim matching"""
    config_file = tmp_path / "auth.cfg"
    with open(config_file, "w+", encoding="utf-8") as fd:
        fd.write(
            """
[auth_jwt_test]
algorithm=HS256
sign=true
client_types=agent,compiler
key=eciwliGyqECVmXtIkNpfVrtBLutZiITZKSKYhogeHMM
expire=0
issuer=https://localhost:8888/
audience=https://localhost:8888/
claims=
    prod in environment
    type is dc
"""
        )

    # Make sure the config starts from a clean slate
    auth.AuthJWTConfig.reset()
    config.Config.load_config(str(config_file))

    # load and parse
    cfg = auth.AuthJWTConfig.get("test")
    assert cfg

    assert auth.check_custom_claims({"environment": ["prod"], "type": "dc"}, cfg.claims)
    assert not auth.check_custom_claims({"environment": ["prod"], "type": ["dc"]}, cfg.claims)
    assert auth.check_custom_claims({"environment": ["prod"], "type": "dc", "other": "test"}, cfg.claims)
    assert not auth.check_custom_claims({"environment": ["prod"]}, cfg.claims)
    assert auth.check_custom_claims({"environment": ["prod"]}, [])


async def test_customer_header_user(tmp_path: pathlib.Path, server: Server) -> None:
    """Test using custom header and users"""
    port = config.Config.get("client_rest_transport", "port")
    config_file = tmp_path / "auth.cfg"
    with open(config_file, "w+", encoding="utf-8") as fd:
        fd.write(
            """
[server]
auth=true
auth_additional_header=Jwt-Assertion

[auth_jwt_test]
algorithm=HS256
sign=true
client_types=agent,compiler
key=eciwliGyqECVmXtIkNpfVrtBLutZiITZKSKYhogeHMM
expire=0
issuer=https://localhost:8888/
audience=https://localhost:8888/

[auth_jwt_name]
algorithm=HS256
client_types=agent,compiler
key=eciwliGyqECVmXtIkNpfVrtBLutZiITZKSKYhogeHMM
expire=0
issuer=https://example.com:8888/
audience=abcdef
jwt_username_claim=name
"""
        )

    # Make sure the config starts from a clean slate
    config.Config.load_config(str(config_file))
    config.Config.set("client_rest_transport", "port", str(port))

    # load and parse
    cfg = auth.AuthJWTConfig.get("test")
    assert cfg

    cfg = auth.AuthJWTConfig.get("name")
    assert cfg

    # Test authentication with custom token header
    payload: dict[str, object] = {
        "iss": cfg.issuer,
        "aud": [cfg.audience],
        const.INMANTA_URN + "ct": "api",
        "name": "test-user",
    }
    token = jwt.encode(payload=payload, key=cfg.key, algorithm=cfg.algo)

    def get() -> requests.Response:
        return requests.get(f"http://localhost:{port}/api/v2/current_user", headers={"Jwt-Assertion": token})

    loop = asyncio.get_event_loop()
    response = await loop.run_in_executor(None, get)
    response.raise_for_status()

    assert response.json()["data"]["username"] == "test-user"



"""
Copyright 2019 Inmanta

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Contact: code@inmanta.com

Module defining the v2 rest api
"""

import datetime
import uuid
from collections.abc import Mapping, Sequence
from typing import Literal, Optional, Union

import inmanta.types
from inmanta.const import AgentAction, AllAgentAction, ApiDocsFormat, Change, ClientType, ParameterSource, ResourceState
from inmanta.data import model
from inmanta.data.model import DataBaseReport, GetSourceCodeResponse, LinkedDiscoveredResource, PipConfig
from inmanta.protocol import methods
from inmanta.protocol.common import ReturnValue
from inmanta.protocol.decorators import typedmethod
from inmanta.protocol.openapi.model import OpenAPI
from inmanta.types import PrimitiveTypes, ResourceIdStr


@typedmethod(
    path="/version/partial",
    operation="PUT",
    arg_options=methods.ENV_OPTS,
    client_types=[ClientType.compiler],
    api_version=2,
    varkw=True,
)
def put_partial(
    tid: uuid.UUID,
    resource_state: Optional[Mapping[ResourceIdStr, Literal[ResourceState.available, ResourceState.undefined]]] = None,
    unknowns: Optional[Sequence[Mapping[str, PrimitiveTypes]]] = None,
    resource_sets: Optional[Mapping[ResourceIdStr, Optional[str]]] = None,
    removed_resource_sets: Optional[Sequence[str]] = None,
    pip_config: Optional[PipConfig] = None,
    **kwargs: object,  # bypass the type checking for the resources and version_info argument
) -> ReturnValue[int]:
    """
    Store a new version of the configuration model after a partial recompile. The partial is applied on top of the latest
    version. Dynamically acquires a new version and serializes concurrent calls. Python code for the new version is copied
    from the base version.

    Concurrent put_partial calls are safe from race conditions provided that their resource sets are disjunct. A put_version
    call concurrent with a put_partial is not guaranteed to be safe. It is the caller's responsibility to appropriately
    serialize them with respect to one another. The caller must ensure the reserve_version + put_version operation is atomic
    with respect to put_partial. In other words, put_partial must not be called in the window between reserve_version and
    put_version. If not respected, either the full or the partial export might be immediately stale, and future exports will
    only be applied on top of the non-stale one.

    :param tid: The id of the environment
    :param resource_state: A dictionary with the initial const.ResourceState per resource id. The ResourceState should be set
                           to undefined when the resource depends on an unknown or available when it doesn't.
    :param unknowns: A list of unknown parameters that caused the model to be incomplete
    :param resource_sets: a dictionary describing which resources belong to which resource set
    :param removed_resource_sets: a list of resource_sets that should be deleted from the model
    :param **kwargs: The following arguments are supported:
              * resources: a list of resource objects. Since the version is not known yet resource versions should be set to 0.
              * version_info: Model version information
    :param pip_config: Pip config used by this version
    :return: The newly stored version number.
    """


# Method for working with projects
@typedmethod(path="/project", operation="PUT", client_types=[ClientType.api], api_version=2)
def project_create(name: str, project_id: Optional[uuid.UUID] = None) -> model.Project:
    """
    Create a new project

    :param name: The name of the project
    :param project_id: A unique uuid, when it is not provided the server generates one
    """


@typedmethod(path="/project/<id>", operation="POST", client_types=[ClientType.api], api_version=2)
def project_modify(id: uuid.UUID, name: str) -> model.Project:
    """
    Rename the given project

    :param id: The id of the project to be modified.
    :param name: The new name for the project. This string value will replace the current name of the project.
    """


@typedmethod(path="/project/<id>", operation="DELETE", client_types=[ClientType.api], api_version=2)
def project_delete(id: uuid.UUID) -> None:
    """
    Delete the given project and all related data

    :param id: The id of the project to be deleted.
    """


@typedmethod(path="/project", operation="GET", client_types=[ClientType.api], api_version=2)
def project_list(environment_details: bool = False) -> list[model.Project]:
    """
    Returns a list of projects ordered alphabetically by name. The environments within each project are also sorted by name.

    :param environment_details: Whether to include the icon and description of the environments in the results
    """


@typedmethod(path="/project/<id>", operation="GET", client_types=[ClientType.api], api_version=2)
def project_get(id: uuid.UUID, environment_details: bool = False) -> model.Project:
    """
    Get a project and a list of the environments under this project

    :param id: The id for which the project's details are being requested.
    :param environment_details: Whether to include the icon and description of the environments in the results
    """


# Methods for working with environments
@typedmethod(path="/environment", operation="PUT", client_types=[ClientType.api], api_version=2)
def environment_create(
    project_id: uuid.UUID,
    name: str,
    repository: Optional[str] = None,
    branch: Optional[str] = None,
    environment_id: Optional[uuid.UUID] = None,
    description: str = "",
    icon: str = "",
) -> model.Environment:
    """
    Create a new environment

    :param project_id: The id of the project this environment belongs to
    :param name: The name of the environment. The name should be unique for each project.
    :param repository: The url (in git form) of the repository
    :param branch: The name of the branch in the repository
    :param environment_id: A unique environment id, if none an id is allocated by the server
    :param description: The description of the environment, maximum 255 characters
    :param icon: The data-url of the icon of the environment. It should follow the pattern `<mime-type>;base64,<image>`, where
                 <mime-type> is one of: 'image/png', 'image/jpeg', 'image/webp', 'image/svg+xml', and <image> is the image in
                 the format matching the specified mime-type, and base64 encoded.
                 The length of the whole string should be maximum 64 kb.

    :raises BadRequest: When the parameters supplied are not valid.
    """


@typedmethod(path="/environment/<id>", operation="POST", client_types=[ClientType.api], api_version=2)
def environment_modify(
    id: uuid.UUID,
    name: str,
    repository: Optional[str] = None,
    branch: Optional[str] = None,
    project_id: Optional[uuid.UUID] = None,
    description: Optional[str] = None,
    icon: Optional[str] = None,
) -> model.Environment:
    """
    Modify the given environment
    The optional parameters that are unspecified will be left unchanged by the update.

    :param id: The id of the environment
    :param name: The name of the environment
    :param repository: The url (in git form) of the repository
    :param branch: The name of the branch in the repository
    :param project_id: The id of the project the environment belongs to
    :param description: The description of the environment, maximum 255 characters
    :param icon: The data-url of the icon of the environment. It should follow the pattern `<mime-type>;base64,<image>` , where
                 <mime-type> is one of: 'image/png', 'image/jpeg', 'image/webp', 'image/svg+xml', and <image> is the image in
                 the format matching the specified mime-type, and base64 encoded.
                 The length of the whole string should be maximum 64 kb.
                 The icon can be removed by setting this parameter to an empty string.

    :raises BadRequest: When the parameters supplied are not valid.
    :raises NotFound: The given environment doesn't exist.
    """


@typedmethod(path="/environment/<id>", operation="DELETE", client_types=[ClientType.api], api_version=2)
def environment_delete(id: uuid.UUID) -> None:
    """
    Delete the given environment and all related data.

    :param id: The uuid of the environment.

    :raises NotFound: The given environment doesn't exist.
    :raises Forbidden: The given environment is protected.
    """


@typedmethod(path="/environment", operation="GET", client_types=[ClientType.api], api_version=2)
def environment_list(details: bool = False) -> list[model.Environment]:
    """
    Returns a list of environments

    :param details: Whether to include the icon and description of the environments in the results
    """


@typedmethod(
    path="/environment/<id>",
    operation="GET",
    client_types=[ClientType.api],
    arg_options={"id": methods.ArgOption(getter=methods.add_env)},
    api_version=2,
)
def environment_get(id: uuid.UUID, details: bool = False) -> model.Environment:
    """
    Get an environment and all versions associated

    :param id: The id of the environment to return
    :param details: Whether to include the icon and description of the environment
    """


@typedmethod(
    path="/actions/environment/halt",
    operation="POST",
    arg_options=methods.ENV_OPTS,
    client_types=[ClientType.api],
    api_version=2,
)
def halt_environment(tid: uuid.UUID) -> None:
    """
    Halt all orchestrator operations for an environment. The environment will enter a state where all agents are paused and
    can not be unpaused. Incoming compile requests will still be queued but compilation will halt. Normal operation can be
    restored using the `resume_environment` endpoint.

    :param tid: The environment id

    :raises NotFound: The given environment doesn't exist.
    """


@typedmethod(
    path="/actions/environment/resume",
    operation="POST",
    arg_options=methods.ENV_OPTS,
    client_types=[ClientType.api],
    api_version=2,
)
def resume_environment(tid: uuid.UUID) -> None:
    """
    Resume all orchestrator operations for an environment. Resumes normal environment operation and unpauses all agents
    that were active when the environment was halted.

    :param tid: The environment id

    :raises NotFound: The given environment doesn't exist.
    """


@typedmethod(
    path="/decommission/<id>",
    operation="DELETE",
    arg_options={"id": methods.ArgOption(getter=methods.convert_environment)},
    client_types=[ClientType.api],
    api_version=2,
)
def environment_clear(id: uuid.UUID) -> None:
    """
    Clear all data from this environment. The environment will be temporarily halted during the decommissioning process.

    :param id: The uuid of the environment.

    :raises NotFound: The given environment doesn't exist.
    :raises Forbidden: The given environment is protected.
    """


# Method for listing and creating auth tokens for an environment that can be used by the agent and compilers
@typedmethod(
    path="/environment_auth",
    operation="POST",
    arg_options=methods.ENV_OPTS,
    client_types=[ClientType.api, ClientType.compiler],
    api_version=2,
)
def environment_create_token(tid: uuid.UUID, client_types: Sequence[str], idempotent: bool = True) -> str:
    """
    Create or get a new token for the given client types. Tokens generated with this call are scoped to the current
    environment.

    :param tid: The environment id
    :param client_types: The client types for which this token is valid (api, agent, compiler)
    :param idempotent: The token should be idempotent, such tokens do not have an expire or issued at set so their
                       value will not change.
    """


# Method for listing/getting/setting/removing settings of an environment. This API is also used by agents to configure
# environments.
@typedmethod(
    path="/environment_settings",
    operation="GET",
    arg_options=methods.ENV_OPTS,
    api=True,
    agent_server=True,
    client_types=[ClientType.api, ClientType.agent, ClientType.compiler],
    api_version=2,
)
def environment_settings_list(tid: uuid.UUID) -> model.EnvironmentSettingsReponse:
    """
    List the settings in the current environment ordered by name alphabetically.

    :param tid: The id of the environment for which the list of settings is being requested.
    """


@typedmethod(
    path="/environment_settings/<id>",
    operation="POST",
    arg_options=methods.ENV_OPTS,
    api=True,
    agent_server=True,
    client_types=[ClientType.api, ClientType.agent, ClientType.compiler],
    api_version=2,
)
def environment_settings_set(tid: uuid.UUID, id: str, value: model.EnvSettingType) -> ReturnValue[None]:
    """
    Set a specific setting in an environment's configuration.

    :param tid: The id of the environment where the setting is to be set or updated.
    :param id: The id of the setting to be set or updated.
    :param value: The new value for the setting.

    """


@typedmethod(
    path="/environment_settings/<id>",
    operation="GET",
    arg_options=methods.ENV_OPTS,
    api=True,
    agent_server=True,
    client_types=[ClientType.api, ClientType.agent],
    api_version=2,
)
def environment_setting_get(tid: uuid.UUID, id: str) -> model.EnvironmentSettingsReponse:
    """
    Retrieve a specific setting from an environment's configuration.

    :param tid: The id of the environment from which the setting is being retrieved.
    :param id: The id of the setting to be retrieved.

    """


@typedmethod(
    path="/environment_settings/<id>",
    operation="DELETE",
    arg_options=methods.ENV_OPTS,
    api=True,
    agent_server=True,
    client_types=[ClientType.api, ClientType.agent],
    api_version=2,
)
def environment_setting_delete(tid: uuid.UUID, id: str) -> ReturnValue[None]:
    """
    Reset the given setting to its default value.

    :param tid: The id of the environment from which the setting is to be deleted.
    :param id: The identifier of the setting to be deleted.
    """


@typedmethod(
    path="/protected_environment_settings",
    operation="POST",
    arg_options=methods.ENV_OPTS,
    api=True,
    client_types=[ClientType.compiler],
    api_version=2,
)
def protected_environment_settings_set_batch(
    tid: uuid.UUID, settings: dict[str, model.EnvSettingType], protected_by: model.ProtectedBy
) -> None:
    """
    Set the values for the given environment settings and mark them as protected.
    If any other environment setting in that environment was protected for the same reason,
    its protected status will be cancelled.

    :param tid: The id of the environment from which to update the environment settings.
    :param settings: The values for each environment setting that has to be updated.
    :param protected_by: The reason why the environment settings must be protected.
    """


@typedmethod(
    path="/reserve_version", operation="POST", arg_options=methods.ENV_OPTS, client_types=[ClientType.compiler], api_version=2
)
def reserve_version(tid: uuid.UUID) -> int:
    """
    Reserve a version number in this environment.

    :param tid: The id of the environment in which the version number is to be reserved.
    """


@typedmethod(path="/docs", operation="GET", client_types=[ClientType.api], api_version=2, token_param="token")
def get_api_docs(
    format: Optional[ApiDocsFormat] = ApiDocsFormat.swagger, token: str | None = None
) -> ReturnValue[Union[OpenAPI, str]]:
    """
    Get the OpenAPI definition of the API

    :param format: Use 'openapi' to get the schema in json format, leave empty or use 'swagger' to get the Swagger-UI view
    :param token: If provided, use this token to authorize the request instead of the value from the authorization header.
    """


@typedmethod(
    path="/scheduler_status",
    operation="GET",
    arg_options=methods.ENV_OPTS,
    client_types=[ClientType.api],
    api_version=2,
)
def get_scheduler_status(tid: uuid.UUID) -> model.SchedulerStatusReport:
    """
    Inspect the scheduler state from the given environment.

    :param tid: The id of the environment in which to inspect the scheduler.
    :raise NotFound: No scheduler is running. For example because the environment is halted.
    """


@typedmethod(
    path="/scheduler/state",
    operation="GET",
    server_agent=True,
    enforce_auth=False,
    arg_options=methods.ENV_OPTS,
    client_types=[],
    api_version=2,
)
def trigger_get_status(tid: uuid.UUID) -> model.SchedulerStatusReport:
    """
    Get a snapshot of the scheduler state

    :param tid: The id of the environment.
    """


@typedmethod(
    path="/scheduler/timers",
    operation="POST",
    server_agent=True,
    timeout=5,
    arg_options=methods.AGENT_ENV_OPTS,
    client_types=[],
    reply=False,
    enforce_auth=False,
)
def notify_timer_update(tid: uuid.UUID) -> None:
    """
    Notify the scheduler of a change in the timer settings
    """


@typedmethod(
    path="/executors/remove_venvs",
    operation="DELETE",
    server_agent=True,
    timeout=5,
    arg_options=methods.AGENT_ENV_OPTS,
    client_types=[],
    reply=False,
    enforce_auth=False,
)
def remove_executor_venvs() -> None:
    """
    Remove all the Python virtual environments.
    """


@typedmethod(
    path="/agent/<name>/<action>", operation="POST", arg_options=methods.ENV_OPTS, client_types=[ClientType.api], api_version=2
)
def agent_action(tid: uuid.UUID, name: str, action: AgentAction) -> None:
    """
    Execute an action on an agent

    :param tid: The environment this agent is defined in.
    :param name: The name of the agent.
    :param action: The type of action that should be executed on an agent.
                    Pause and unpause can only be used when the environment is not halted,
                    while the on_resume actions can only be used when the environment is halted.
                    * pause: A paused agent cannot execute any deploy operations.
                    * unpause: A unpaused agent will be able to execute deploy operations.
                    * keep_paused_on_resume: The agent will still be paused when the environment is resumed
                    * unpause_on_resume: The agent will be unpaused when the environment is resumed.

    :raises Forbidden: The given environment has been halted and the action is pause/unpause,
                        or the environment is not halted and the action is related to the on_resume behavior
    """


@typedmethod(
    path="/agents/<action>", operation="POST", arg_options=methods.ENV_OPTS, client_types=[ClientType.api], api_version=2
)
def all_agents_action(tid: uuid.UUID, action: AllAgentAction) -> None:
    """
    Execute an action on all agents in the given environment.

    :param tid: The environment of the agents.
    :param action: The type of action that should be executed on the agents.

      Pause and unpause can only be used when the environment is not halted,
      while the on_resume actions can only be used when the environment is halted.

      * pause: A paused agent cannot execute any deploy operations.
      * unpause: A unpaused agent will be able to execute deploy operations.
      * keep_paused_on_resume: The agents will still be paused when the environment is resumed
      * unpause_on_resume: The agents will be unpaused when the environment is resumed
      * remove_all_agent_venvs: Remove all agent venvs in the given environment. During this
         process the agent operations in that environment are temporarily
         suspended. The removal of the agent venvs will happen asynchronously
         with respect to this API call. It might take a long time until the
         venvs are actually removed, because all executing agent operations
         will be allowed to finish first. As such, a long deploy operation
         might delay the removal of the venvs considerably.

    :raises Forbidden: The given environment has been halted and the action is pause/unpause,
                        or the environment is not halted and the action is related to the on_resume behavior
    """


@typedmethod(path="/agents", operation="GET", arg_options=methods.ENV_OPTS, client_types=[ClientType.api], api_version=2)
def get_agents(
    tid: uuid.UUID,
    limit: Optional[int] = None,
    start: Optional[Union[datetime.datetime, bool, str]] = None,
    end: Optional[Union[datetime.datetime, bool, str]] = None,
    first_id: Optional[str] = None,
    last_id: Optional[str] = None,
    filter: Optional[Mapping[str, Sequence[str]]] = None,
    sort: str = "name.asc",
) -> list[model.Agent]:
    """
    Get all of the agents in the given environment

    :param tid: The id of the environment the agents should belong to.
    :param limit: Limit the number of agents that are returned.
    :param start: The lower limit for the order by column (exclusive).
    :param first_id: The name to use as a continuation token for paging, in combination with the 'start' value,
        because the order by column might contain non-unique values.
    :param last_id: The name to use as a continuation token for paging, in combination with the 'end' value,
        because the order by column might contain non-unique values.
        Only one of 'start' and 'end' should be specified at the same time.
    :param end: The upper limit for the order by column (exclusive).
        Only one of 'start' and 'end' should be specified at the same time.
    :param filter: Filter the list of returned agents.
        Filtering by 'name', 'process_name' and 'status' is supported.
    :param sort: Return the results sorted according to the parameter value.
        Sorting by 'name', 'process_name', 'status', 'paused' and 'last_failover' is supported.
        The following orders are supported: 'asc', 'desc'
    :return: A list of all matching agents
    :raise NotFound: This exception is raised when the referenced environment is not found
    :raise BadRequest: When the parameters used for filtering, sorting or paging are not valid
    """


@typedmethod(
    path="/agents/process/<id>", operation="GET", arg_options=methods.ENV_OPTS, client_types=[ClientType.api], api_version=2
)
def get_agent_process_details(tid: uuid.UUID, id: uuid.UUID, report: bool = False) -> model.AgentProcess:
    """
    Get the details of an agent process

    :param tid: Id of the environment
    :param id: The id of the specific agent process
    :param report: Whether to include a report from the agent or not
    :return: The details of an agent process
    :raise NotFound: This exception is raised when the referenced environment or agent process is not found
    """


@typedmethod(
    path="/db_status", api=False, server_agent=True, enforce_auth=False, operation="POST", client_types=[], api_version=2
)
def get_db_status() -> DataBaseReport:
    """
    Get a report of the DB connection pool status
    """


@typedmethod(
    path="/compiledata/<id>",
    operation="GET",
    client_types=[ClientType.api],
    api_version=2,
)
def get_compile_data(id: uuid.UUID) -> Optional[model.CompileData]:
    """
    Get the compile data for the given compile request.

    :param id: The id of the compile.
    """


@typedmethod(
    path="/resource_actions", operation="GET", arg_options=methods.ENV_OPTS, client_types=[ClientType.api], api_version=2
)
def get_resource_actions(
    tid: uuid.UUID,
    resource_type: Optional[str] = None,
    agent: Optional[str] = None,
    attribute: Optional[str] = None,
    attribute_value: Optional[str] = None,
    log_severity: Optional[str] = None,
    limit: Optional[int] = 0,
    action_id: Optional[uuid.UUID] = None,
    first_timestamp: Optional[datetime.datetime] = None,
    last_timestamp: Optional[datetime.datetime] = None,
    exclude_changes: Optional[Sequence[Change]] = None,
) -> ReturnValue[list[model.ResourceAction]]:
    """
    Return resource actions matching the search criteria.

    :param tid: The id of the environment this resource belongs to
    :param resource_type: The resource entity type that should be queried
    :param agent: Agent name that is used to filter the results
    :param attribute: Attribute name used for filtering
    :param attribute_value: Attribute value used for filtering. Attribute and attribute value should be supplied together.
    :param log_severity: Only include ResourceActions which have a log message with this severity.
    :param limit: Limit the number of resource actions included in the response, up to 1000
    :param action_id: Start the query from this action_id.
            To be used in combination with either the first or last timestamp.
    :param first_timestamp: Limit the results to resource actions that started later
            than the value of this parameter (exclusive)
    :param last_timestamp: Limit the results to resource actions that started earlier
            than the value of this parameter (exclusive).
            Only the first_timestamp or last_timestamp parameter should be supplied
    :param exclude_changes: only return ResourceActions where the change type is different from the one in this list.
    :return: The list of matching Resource Actions.
            The order is ascending if first_timestamp is provided, otherwise descending.
            If a limit is specified, also return links to the next and previous pages.
            The "next" page refers to actions that started earlier, while the "prev" page refers to actions that started later.


    :raises BadRequest: When the supplied parameters are not valid.

    """


# No pagination support is provided for this endpoint because there is no elegant way to page the output of this endpoint.
@typedmethod(
    path="/resource/<rvid>/events",
    operation="GET",
    arg_options={**methods.ENV_OPTS, **methods.RVID_OPTS},
    agent_server=True,
    client_types=[ClientType.agent],
    api_version=2,
)
def get_resource_events(
    tid: uuid.UUID,
    rvid: inmanta.types.ResourceVersionIdStr,
    exclude_change: Optional[Change] = None,
) -> dict[inmanta.types.ResourceIdStr, list[model.ResourceAction]]:
    """
    Return relevant events for a resource, i.e. all deploy actions for each of its dependencies since this resources' last
    successful deploy or all deploy actions if this resources hasn't been deployed before. The resource actions are sorted in
    descending order according to their started timestamp. If exclude_change is set, exclude all resource actions with this
    specific type of change.

    This method searches through all versions of this resource.
    This method should only be called when a deploy is in progress.

    :param tid: The id of the environment this resource belongs to
    :param rvid: The id of the resource to get events for.
    :param exclude_change: Exclude all resource actions with this specific type of change.
    :raises BadRequest: When this endpoint in called while the resource with the given resource version is not
                        in the deploying state.
    """


@typedmethod(
    path="/resource/<rvid>/did_dependency_change",
    operation="GET",
    arg_options={**methods.ENV_OPTS, **methods.RVID_OPTS},
    agent_server=True,
    client_types=[ClientType.agent],
    api_version=2,
)
def resource_did_dependency_change(
    tid: uuid.UUID,
    rvid: inmanta.types.ResourceVersionIdStr,
) -> bool:
    """
    Returns True iff this resources' events indicate a change in its dependencies since the resource's last deployment.

    This method searches through all versions of this resource.
    This method should only be called when a deploy is in progress.

    :param tid: The id of the environment this resource belongs to
    :param rvid: The id of the resource.
    :raises BadRequest: When this endpoint in called while the resource with the given resource version is not
                        in the deploying state.
    """


@typedmethod(path="/resource", operation="GET", arg_options=methods.ENV_OPTS, client_types=[ClientType.api], api_version=2)
def resource_list(
    tid: uuid.UUID,
    limit: Optional[int] = None,
    first_id: Optional[inmanta.types.ResourceVersionIdStr] = None,
    last_id: Optional[inmanta.types.ResourceVersionIdStr] = None,
    start: Optional[str] = None,
    end: Optional[str] = None,
    filter: Optional[Mapping[str, Sequence[str]]] = None,
    sort: str = "resource_type.desc",
    deploy_summary: bool = False,
) -> list[model.LatestReleasedResource]:
    """
    :param tid: The id of the environment this resource belongs to
    :param limit: Limit the number of instances that are returned
    :param first_id: The resource_version_id to use as a continuation token for paging, in combination with the 'start' value,
            because the order by column might contain non-unique values
    :param last_id: The resource_version_id to use as a continuation token for paging, in combination with the 'end' value,
            because the order by column might contain non-unique values
    :param start: The lower limit for the order by column (exclusive).
                Only one of 'start' and 'end' should be specified at the same time.
    :param end: The upper limit for the order by column (exclusive).
                Only one of 'start' and 'end' should be specified at the same time.
    :param filter: Filter the list of returned resources.
                Filters should be specified with the syntax `?filter.<filter_key>=value`, for example `?filter.status=deployed`
                It's also possible to provide multiple values for the same filter, in this case resources are returned,
                if they match any of these filter values.
                For example: `?filter.status=deployed&filter.status=available` returns instances with either of the statuses
                deployed or available.
                Multiple different filters narrow the results however (they are treated as an 'AND' operator).
                For example `filter.status=deployed&filter.agent=internal` returns resources
                with 'deployed' status, where the 'agent' is set to 'internal_agent'.
                The following options are available:
                agent: filter by the agent of the resource
                resource_type: filter by the type of the resource
                resource_id_value: filter by the attribute values of the resource
                status: filter by the current status of the resource.
                For status filters it's also possible to invert the condition with '!'.
                The inverted status filter works as a separate filter from the normal status filter.
                For example: `filter.status=!orphaned` will return all the resources that are not in 'orphaned' state.
                If multiple values are provided to the inverted filter, resources are returned if they don't match
                any of the filter values.
                For example: `?filter.status=!deployed&filter.status=!available`
                returns all instances except those whose status is deployed or available.
                The values for the 'agent', 'resource_type' and 'value' filters are matched partially.
    :param sort: Return the results sorted according to the parameter value.
                It should follow the pattern `<attribute_to_sort_by>.<order>`, for example `resource_type.desc`
                (case insensitive).
                The following sorting attributes are supported: 'resource_type', 'agent', 'resource_id_value', 'status'.
                The following orders are supported: 'asc', 'desc'
    :param deploy_summary: If set to true, returns a summary of the deployment status of the resources in the environment
                           in the metadata, describing how many resources are in each state as well as the total number
                           of resources. The summary does not take into account the current filters or paging parameters.
                           Orphaned resources are not included in the summary
    :return: A list of all matching released resources
    :raise NotFound: This exception is raised when the referenced environment is not found
    :raise BadRequest: When the parameters used for filtering, sorting or paging are not valid
    """


@typedmethod(
    path="/resource/<rid>", operation="GET", arg_options=methods.ENV_OPTS, client_types=[ClientType.api], api_version=2
)
def resource_details(tid: uuid.UUID, rid: inmanta.types.ResourceIdStr) -> model.ReleasedResourceDetails:
    """
    :param tid: The id of the environment from which the resource's details are being requested.
    :param rid: The unique identifier (ResourceIdStr) of the resource. This value specifies the particular resource
                for which detailed information is being requested.

    :return: The details of the latest released version of a resource
    :raise NotFound: This exception is raised when the referenced environment or resource is not found
    """


@typedmethod(
    path="/resource/<rid>/history", operation="GET", arg_options=methods.ENV_OPTS, client_types=[ClientType.api], api_version=2
)
def resource_history(
    tid: uuid.UUID,
    rid: inmanta.types.ResourceIdStr,
    limit: Optional[int] = None,
    first_id: Optional[str] = None,
    last_id: Optional[str] = None,
    start: Optional[datetime.datetime] = None,
    end: Optional[datetime.datetime] = None,
    sort: str = "date.desc",
) -> list[model.ResourceHistory]:
    """
    :param tid: The id of the environment this resource belongs to
    :param rid: The id of the resource
    :param limit: Limit the number of instances that are returned
    :param first_id: The attribute_hash to use as a continuation token for paging, in combination with the 'start' value,
            because the order by column might contain non-unique values
    :param last_id: The attribute_hash to use as a continuation token for paging, in combination with the 'end' value,
            because the order by column might contain non-unique values
    :param start: The lower limit for the order by column (exclusive).
                Only one of 'start' and 'end' should be specified at the same time.
    :param end: The upper limit for the order by column (exclusive).
                Only one of 'start' and 'end' should be specified at the same time.
    :param sort: Return the results sorted according to the parameter value.
                It should follow the pattern `<attribute_to_sort_by>.<order>`, for example `date.desc`
                (case insensitive).
                Sorting by `date` is supported.
                The following orders are supported: 'asc', 'desc'
    :return: The history of a resource, according to its attributes
    :raise NotFound: This exception is raised when the referenced environment is not found
    :raise BadRequest: When the parameters used for sorting or paging are not valid
    """


@typedmethod(
    path="/resource/<rid>/logs", operation="GET", arg_options=methods.ENV_OPTS, client_types=[ClientType.api], api_version=2
)
def resource_logs(
    tid: uuid.UUID,
    rid: inmanta.types.ResourceIdStr,
    limit: Optional[int] = None,
    start: Optional[datetime.datetime] = None,
    end: Optional[datetime.datetime] = None,
    filter: Optional[Mapping[str, Sequence[str]]] = None,
    sort: str = "timestamp.desc",
) -> list[model.ResourceLog]:
    """
    Get the logs of a specific resource.

    :param tid: The id of the environment this resource belongs to
    :param rid: The id of the resource
    :param limit: Limit the number of instances that are returned
    :param start: The lower limit for the order by column (exclusive). Only one of 'start' and 'end' should be specified at
        the same time.
    :param end: The upper limit for the order by column (exclusive). Only one of 'start' and 'end' should be specified at the
        same time.
    :param filter: Filter the list of returned logs.
        Filters should be specified with the syntax `?filter.<filter_key>=value`, for example `?filter.minimal_log_level=INFO`.
        It's also possible to provide multiple values for the same filter, in this case resources are returned, if they match
        any of these filter values.

        For example: `?filter.action=pull&filter.action=deploy` returns logs with either of the actions pull or deploy.
        Multiple different filters narrow the results however (they are treated as an 'AND' operator).
        For example `filter.minimal_log_level=INFO&filter.action=deploy` returns logs with 'deploy' action, where the
        'log_level' is at least 'INFO'.

        The following options are available:
            * action: filter by the action of the log

            * timestamp: return the logs matching the timestamp constraints. Valid constraints are of the form
              "<lt|le|gt|ge>:<x>". The expected format is YYYY-MM-DDTHH:mm:ss.ssssss, so an ISO-8601 datetime string,
              in UTC timezone.

            For example: `?filter.timestamp=ge:2021-08-18T09:21:30.568353&filter.timestamp=lt:2021-08-18T10:21:30.568353`.
            Multiple constraints can be specified, in which case only log messages that match all constraints will be
            returned.

            * message: filter by the content of the log messages. Partial matches are allowed. (case-insensitive)

            * minimal_log_level: filter by the log level of the log messages. The filter specifies the minimal level,
              so messages with either this level, or a higher severity level are going to be included in the result.

            For example, for `filter.minimal_log_level=INFO`, the log messages with level `INFO, WARNING, ERROR, CRITICAL`
            all match the query.

    :param sort: Return the results sorted according to the parameter value. It should follow the pattern
        `<attribute_to_sort_by>.<order>`, for example `timestamp.desc` (case insensitive). Only sorting by `timestamp` is
        supported. The following orders are supported: 'asc', 'desc'

    :return: A list of all matching resource logs
    :raise NotFound: This exception is raised when the referenced environment is not found
    :raise BadRequest: When the parameters used for filtering, sorting or paging are not valid
    """


@typedmethod(
    path="/resource/<rid>/facts", operation="GET", arg_options=methods.ENV_OPTS, client_types=[ClientType.api], api_version=2
)
def get_facts(tid: uuid.UUID, rid: inmanta.types.ResourceIdStr) -> list[model.Fact]:
    """
    Get the facts related to a specific resource. The results are sorted alphabetically by name.
    :param tid: The id of the environment
    :param rid: Id of the resource
    :return: The facts related to this resource
    :raise NotFound: This status code is returned when the referenced environment is not found
    """


@typedmethod(
    path="/resource/<rid>/facts/<id>",
    operation="GET",
    arg_options=methods.ENV_OPTS,
    client_types=[ClientType.api],
    api_version=2,
)
def get_fact(tid: uuid.UUID, rid: inmanta.types.ResourceIdStr, id: uuid.UUID) -> model.Fact:
    """
    Get one specific fact
    :param tid: The id of the environment
    :param rid: The id of the resource
    :param id: The id of the fact
    :return: A specific fact corresponding to the id
    :raise NotFound: This status code is returned when the referenced environment or fact is not found
    """


@typedmethod(path="/compilereport", operation="GET", arg_options=methods.ENV_OPTS, client_types=[ClientType.api], api_version=2)
def get_compile_reports(
    tid: uuid.UUID,
    limit: Optional[int] = None,
    first_id: Optional[uuid.UUID] = None,
    last_id: Optional[uuid.UUID] = None,
    start: Optional[datetime.datetime] = None,
    end: Optional[datetime.datetime] = None,
    filter: Optional[Mapping[str, Sequence[str]]] = None,
    sort: str = "requested.desc",
) -> list[model.CompileReport]:
    """
    Get the compile reports from an environment.

    The returned compile report objects may carry links to other objects, e.g. a service instance.
    The full list of supported links can be found :ref:`here <api_self_referencing_links>`.

    :param tid: The id of the environment
    :param limit: Limit the number of instances that are returned
    :param first_id: The id to use as a continuation token for paging, in combination with the 'start' value,
            because the order by column might contain non-unique values
    :param last_id: The id to use as a continuation token for paging, in combination with the 'end' value,
            because the order by column might contain non-unique values
    :param start: The lower limit for the order by column (exclusive).
                Only one of 'start' and 'end' should be specified at the same time.
    :param end: The upper limit for the order by column (exclusive).
                Only one of 'start' and 'end' should be specified at the same time.
    :param filter: Filter the list of returned compile reports.
                Filters should be specified with the syntax `?filter.<filter_key>=value`,
                for example `?filter.success=True`
                It's also possible to provide multiple values for the same filter, in this case resources are returned,
                if they match any of these filter values.
                For example: `?filter.requested=ge:2021-08-18T09:21:30.568353&filter.requested=lt:2021-08-18T10:21:30.568353`
                returns compile reports that were requested between the specified dates.
                Multiple different filters narrow the results however (they are treated as an 'AND' operator).
                For example `?filter.success=True&filter.completed=True` returns compile reports
                that are completed and successful.
                The following options are available:
                success: whether the compile was successful or not
                started: whether the compile has been started or not
                completed: whether the compile has been completed or not

                requested: return the logs matching the timestamp constraints. Valid constraints are of the form
                    "<lt|le|gt|ge>:<x>". The expected format is YYYY-MM-DDTHH:mm:ss.ssssss, so an ISO-8601 datetime string,
                    in UTC timezone. Specifying microseconds is optional. For example:
                    `?filter.requested=ge:2021-08-18T09:21:30.568353&filter.requested=lt:2021-08-18T10:21:30`.
                    Multiple constraints can be specified, in which case only compile reports that match all constraints will be
                    returned.
    :param sort: Return the results sorted according to the parameter value.
                It should follow the pattern `?sort=<attribute_to_sort_by>.<order>`, for example `?sort=requested.desc`
                (case insensitive).
                Only sorting by the `requested` timestamp is supported.
                The following orders are supported: 'asc', 'desc'
    :return: A list of all matching compile reports
    :raise NotFound: This exception is raised when the referenced environment is not found
    :raise BadRequest: When the parameters used for filtering, sorting or paging are not valid
    """


@typedmethod(
    path="/compilereport/<id>", operation="GET", arg_options=methods.ENV_OPTS, client_types=[ClientType.api], api_version=2
)
def compile_details(tid: uuid.UUID, id: uuid.UUID) -> model.CompileDetails:
    """
    The returned compile details object may carry links to other objects, e.g. a service instance.
    The full list of supported links can be found :ref:`here <api_self_referencing_links>`.

    :param tid: The id of the environment in which the compilation process occurred.
    :param id: The id of the compile for which the details are being requested.

    :return: The details of a compile
    :raise NotFound: This exception is raised when the referenced environment or compile is not found
    """


@typedmethod(path="/desiredstate", operation="GET", arg_options=methods.ENV_OPTS, client_types=[ClientType.api], api_version=2)
def list_desired_state_versions(
    tid: uuid.UUID,
    limit: Optional[int] = None,
    start: Optional[int] = None,
    end: Optional[int] = None,
    filter: Optional[Mapping[str, Sequence[str]]] = None,
    sort: str = "version.desc",
) -> list[model.DesiredStateVersion]:
    """
    Get the desired state versions from an environment.

    :param tid: The id of the environment
    :param limit: Limit the number of versions that are returned
    :param start: The lower limit for the order by column (exclusive).
                Only one of 'start' and 'end' should be specified at the same time.
    :param end: The upper limit for the order by column (exclusive).
                Only one of 'start' and 'end' should be specified at the same time.
    :param filter: Filter the list of returned desired state versions.
                Filtering by 'version' range, 'date' range, 'status' and `released` are supported.
    :param sort: Return the results sorted according to the parameter value.
                Only sorting by 'version' is supported.
                The following orders are supported: 'asc', 'desc'
    :return: A list of all matching compile reports
    :raise NotFound: This exception is raised when the referenced environment is not found
    :raise BadRequest: When the parameters used for filtering, sorting or paging are not valid
    """


@typedmethod(
    path="/desiredstate/<version>/promote",
    operation="POST",
    arg_options=methods.ENV_OPTS,
    client_types=[ClientType.api],
    api_version=2,
)
def promote_desired_state_version(
    tid: uuid.UUID, version: int, trigger_method: Optional[model.PromoteTriggerMethod] = None
) -> None:
    """
    Promote a desired state version, making it the active version in the environment.

    :param tid: The id of the environment
    :param version: The number of the version to promote
    :param trigger_method: [DEPRECATED] This argument is ignored.
    """


@typedmethod(
    path="/desiredstate/<version>",
    operation="GET",
    arg_options=methods.ENV_OPTS,
    client_types=[ClientType.api],
    api_version=2,
)
def get_resources_in_version(
    tid: uuid.UUID,
    version: int,
    limit: Optional[int] = None,
    first_id: Optional[inmanta.types.ResourceVersionIdStr] = None,
    last_id: Optional[inmanta.types.ResourceVersionIdStr] = None,
    start: Optional[str] = None,
    end: Optional[str] = None,
    filter: Optional[Mapping[str, Sequence[str]]] = None,
    sort: str = "resource_type.desc",
) -> list[model.VersionedResource]:
    """
    Get the resources that belong to a specific version.

    :param tid: The id of the environment
    :param version: The version number
    :param limit: Limit the number of resources that are returned
    :param first_id: The resource_version_id to use as a continuation token for paging, in combination with the 'start' value,
            because the order by column might contain non-unique values
    :param last_id: The resource_version_id to use as a continuation token for paging, in combination with the 'end' value,
            because the order by column might contain non-unique values
    :param start: The lower limit for the order by column (exclusive).
                Only one of 'start' and 'end' should be specified at the same time.
    :param end: The upper limit for the order by column (exclusive).
                Only one of 'start' and 'end' should be specified at the same time.
    :param filter: Filter the list of returned resources.
                The following options are available:
                agent: filter by the agent name of the resource
                resource_type: filter by the type of the resource
                resource_id_value: filter by the attribute values of the resource
    :param sort: Return the results sorted according to the parameter value.
                The following sorting attributes are supported: 'resource_type', 'agent', 'resource_id_value'.
                The following orders are supported: 'asc', 'desc'
    :return: A list of all matching resources
    :raise NotFound: This exception is raised when the referenced environment is not found
    :raise BadRequest: When the parameters used for filtering, sorting or paging are not valid
    """


@typedmethod(
    path="/desiredstate/diff/<from_version>/<to_version>",
    operation="GET",
    arg_options=methods.ENV_OPTS,
    client_types=[ClientType.api],
    api_version=2,
)
def get_diff_of_versions(
    tid: uuid.UUID,
    from_version: int,
    to_version: int,
) -> list[model.ResourceDiff]:
    """
    Compare two versions of desired states, and provide the difference between them,
    with regard to their resources and the attributes of these resources.
    Resources that are the same in both versions are not mentioned in the results.

    A resource diff describes whether the resource was 'added', 'modified' or 'deleted',
    and what the values of their attributes were in the versions.
    The values are also returned in a stringified, easy to compare way,
    which can be used to calculate a `git diff`-like summary of the changes.

    :param tid: The id of the environment
    :param from_version: The (lower) version number to compare
    :param to_version: The other (higher) version number to compare
    :return: The resource diffs between from_version and to_version
    :raise NotFound: This exception is raised when the referenced environment or versions are not found
    :raise BadRequest: When the version parameters are not valid
    """


@typedmethod(
    path="/desiredstate/<version>/resource/<rid>",
    operation="GET",
    arg_options=methods.ENV_OPTS,
    client_types=[ClientType.api],
    api_version=2,
)
def versioned_resource_details(
    tid: uuid.UUID, version: int, rid: inmanta.types.ResourceIdStr
) -> model.VersionedResourceDetails:
    """
    :param tid: The id of the environment
    :param version: The version number of the resource
    :param rid: The id of the resource
    :return: The details of a specific version of a resource
    :raise NotFound: This exception is raised when the referenced environment or resource is not found
    """


@typedmethod(
    path="/parameters",
    operation="GET",
    arg_options=methods.ENV_OPTS,
    client_types=[ClientType.api],
    api_version=2,
)
def get_parameters(
    tid: uuid.UUID,
    limit: Optional[int] = None,
    first_id: Optional[uuid.UUID] = None,
    last_id: Optional[uuid.UUID] = None,
    start: Optional[Union[datetime.datetime, str]] = None,
    end: Optional[Union[datetime.datetime, str]] = None,
    filter: Optional[Mapping[str, Sequence[str]]] = None,
    sort: str = "name.asc",
) -> list[model.Parameter]:
    """
    List the parameters in an environment

    :param tid: The id of the environment
    :param limit: Limit the number of parameters that are returned
    :param first_id: The parameter id to use as a continuation token for paging, in combination with the 'start' value,
        because the order by column might contain non-unique values
    :param last_id: The parameter id to use as a continuation token for paging, in combination with the 'end' value,
        because the order by column might contain non-unique values
    :param start: The lower limit for the order by column (exclusive). Only one of 'start' and 'end' should be specified at the
        same time.
    :param end: The upper limit for the order by column (exclusive). Only one of 'start' and 'end' should be specified at the
        same time.
    :param filter: Filter the list of returned parameters.

        The following options are available:
            * name: filter by the name of the parameter
            * source: filter by the source of the parameter
            * updated: filter by the updated time of the parameter
    :param sort: Return the results sorted according to the parameter value.
        The following sorting attributes are supported: 'name', 'source', 'updated'.
        The following orders are supported: 'asc', 'desc'
    :return: A list of all matching parameters
    :raise NotFound: This exception is raised when the referenced environment is not found
    :raise BadRequest: When the parameters used for filtering, sorting or paging are not valid
    """


@typedmethod(
    path="/parameters/<name>",
    operation="PUT",
    arg_options=methods.ENV_OPTS,
    client_types=[ClientType.api, ClientType.compiler, ClientType.agent],
    api_version=2,
)
def set_parameter(
    tid: uuid.UUID,
    name: str,
    source: ParameterSource,
    value: str,
    metadata: Optional[Mapping[str, str]] = None,
    recompile: bool = False,
) -> ReturnValue[model.Parameter]:
    """
    Set a parameter on the server. If the parameter is an tracked unknown, it will trigger a recompile on the server.
    Otherwise, if the value is changed and recompile is true, a recompile is also triggered.

    :param tid: The id of the environment
    :param name: The name of the parameter
    :param source: The source of the parameter.
    :param value: The value of the parameter
    :param metadata: Optional. Metadata about the parameter
    :param recompile: Optional. Whether to trigger a recompile if the value of the parameter changed.
    """


@typedmethod(
    path="/facts",
    operation="GET",
    arg_options=methods.ENV_OPTS,
    client_types=[ClientType.api],
    api_version=2,
)
def get_all_facts(
    tid: uuid.UUID,
    limit: Optional[int] = None,
    first_id: Optional[uuid.UUID] = None,
    last_id: Optional[uuid.UUID] = None,
    start: Optional[str] = None,
    end: Optional[str] = None,
    filter: Optional[Mapping[str, Sequence[str]]] = None,
    sort: str = "name.asc",
) -> list[model.Fact]:
    """
    List the facts in an environment.

    :param tid: The id of the environment
    :param limit: Limit the number of facts that are returned
    :param first_id: The fact id to use as a continuation token for paging, in combination with the 'start' value,
            because the order by column might contain non-unique values
    :param last_id: The fact id to use as a continuation token for paging, in combination with the 'end' value,
            because the order by column might contain non-unique values
    :param start: The lower limit for the order by column (exclusive).
                Only one of 'start' and 'end' should be specified at the same time.
    :param end: The upper limit for the order by column (exclusive).
                Only one of 'start' and 'end' should be specified at the same time.
    :param filter: Filter the list of returned facts.
                The following options are available:
                name: filter by the name of the fact
                resource_id: filter by the resource_id of the fact
                expires: filter on whether the fact expires or not
    :param sort: Return the results sorted according to the parameter value.
                The following sorting attributes are supported: 'name', 'resource_id'.
                The following orders are supported: 'asc', 'desc'
    :return: A list of all matching facts
    :raise NotFound: This exception is raised when the referenced environment is not found
    :raise BadRequest: When the parameters used for filtering, sorting or paging are not valid
    """


@typedmethod(
    path="/facts/<name>",
    operation="PUT",
    arg_options=methods.ENV_OPTS,
    client_types=[ClientType.api, ClientType.compiler, ClientType.agent],
    api_version=2,
)
def set_fact(
    tid: uuid.UUID,
    name: str,
    source: ParameterSource,
    value: str,
    resource_id: str,
    metadata: Optional[Mapping[str, str]] = None,
    recompile: bool = False,
    expires: Optional[bool] = True,
) -> ReturnValue[model.Fact]:
    """
    Set a fact on the server. If the fact is a tracked unknown, it will trigger a recompile on the server.
    Otherwise, if the value is changed and recompile is true, a recompile is also triggered.

    :param tid: The id of the environment
    :param name: The name of the fact
    :param source: The source of the fact
    :param value: The value of the fact
    :param resource_id: The resource this fact belongs to
    :param metadata: Optional. Metadata about the fact
    :param recompile: Optional. Whether to trigger a recompile if the value of the fact changed.
    :param expires: Optional. If the fact should expire or not. By default, facts expire.
    """


# Dryrun related methods


@typedmethod(
    path="/dryrun/<version>", operation="POST", arg_options=methods.ENV_OPTS, client_types=[ClientType.api], api_version=2
)
def dryrun_trigger(tid: uuid.UUID, version: int) -> uuid.UUID:
    """
    Trigger a new dryrun

    :param tid: The id of the environment
    :param version: The version of the configuration model to execute the dryrun for
    :raise NotFound: This exception is raised when the referenced environment or version is not found
    :return: The id of the new dryrun
    """


@typedmethod(
    path="/dryrun/<version>", operation="GET", arg_options=methods.ENV_OPTS, client_types=[ClientType.api], api_version=2
)
def list_dryruns(tid: uuid.UUID, version: int) -> list[model.DryRun]:
    """
    Query a list of dry runs for a specific version

    :param tid: The id of the environment
    :param version: The configuration model version to return dryruns for
    :raise NotFound: This exception is raised when the referenced environment or version is not found
    :return: The list of dryruns for the specified version in descending order by date
    """


@typedmethod(
    path="/dryrun/<version>/<report_id>",
    operation="GET",
    arg_options=methods.ENV_OPTS,
    client_types=[ClientType.api],
    api_version=2,
)
def get_dryrun_diff(tid: uuid.UUID, version: int, report_id: uuid.UUID) -> model.DryRunReport:
    """
    Get the report of a dryrun, describing the changes a deployment would make,
    with the difference between the current and target states provided in a form similar to the desired state diff endpoint.

    :param tid: The id of the environment
    :param version: The version of the configuration model the dryrun belongs to
    :param report_id: The dryrun id to calculate the diff for
    :raise NotFound: This exception is raised when the referenced environment or version is not found
    :return: The dryrun report, with a summary and the list of differences.
    """


@typedmethod(
    path="/notification",
    operation="GET",
    arg_options=methods.ENV_OPTS,
    client_types=[ClientType.api],
    api_version=2,
)
def list_notifications(
    tid: uuid.UUID,
    limit: Optional[int] = None,
    first_id: Optional[uuid.UUID] = None,
    last_id: Optional[uuid.UUID] = None,
    start: Optional[datetime.datetime] = None,
    end: Optional[datetime.datetime] = None,
    filter: Optional[Mapping[str, Sequence[str]]] = None,
    sort: str = "created.desc",
) -> list[model.Notification]:
    """
    List the notifications in an environment.

    The returned notification objects may carry links to other objects, e.g. a compile report. The full list of supported links
    can be found :ref:`here <api_self_referencing_links>`.

    :param tid: The id of the environment
    :param limit: Limit the number of notifications that are returned
    :param first_id: The notification id to use as a continuation token for paging, in combination with the 'start' value,
            because the order by column might contain non-unique values
    :param last_id: The notification id to use as a continuation token for paging, in combination with the 'end' value,
            because the order by column might contain non-unique values
    :param start: The lower limit for the order by column (exclusive).
                Only one of 'start' and 'end' should be specified at the same time.
    :param end: The upper limit for the order by column (exclusive).
                Only one of 'start' and 'end' should be specified at the same time.
    :param filter: Filter the list of returned notifications.
                The following options are available:
                read: Whether the notification was read or not
                cleared: Whether the notification was cleared or not
                severity: Filter by the severity field of the notifications
                title: Filter by the title of the notifications
                message: Filter by the message of the notifications
    :param sort: Return the results sorted according to the parameter value.
                Only sorting by the 'created' date is supported.
                The following orders are supported: 'asc', 'desc'
    :return: A list of all matching notifications
    :raise NotFound: This exception is raised when the referenced environment is not found
    :raise BadRequest: When the parameters used for filtering or paging are not valid
    """


@typedmethod(
    path="/notification/<notification_id>",
    operation="GET",
    arg_options=methods.ENV_OPTS,
    client_types=[ClientType.api],
    api_version=2,
)
def get_notification(
    tid: uuid.UUID,
    notification_id: uuid.UUID,
) -> model.Notification:
    """
    Get a single notification

    :param tid: The id of the environment
    :param notification_id: The id of the notification
    :return: The notification with the specified id
    :raise NotFound: When the referenced environment or notification is not found
    """


@typedmethod(
    path="/notification/<notification_id>",
    operation="PATCH",
    arg_options=methods.ENV_OPTS,
    client_types=[ClientType.api],
    api_version=2,
)
def update_notification(
    tid: uuid.UUID,
    notification_id: uuid.UUID,
    read: Optional[bool] = None,
    cleared: Optional[bool] = None,
) -> model.Notification:
    """
    Update a notification by setting its flags

    :param tid: The id of the environment
    :param notification_id: The id of the notification to update
    :param read: Whether the notification has been read
    :param cleared: Whether the notification has been cleared
    :return: The updated notification
    :raise NotFound: When the referenced environment or notification is not found
    """


@typedmethod(
    path="/code/<version>",
    operation="GET",
    agent_server=True,
    arg_options=methods.ENV_OPTS,
    client_types=[ClientType.agent],
    api_version=2,
)
def get_source_code(tid: uuid.UUID, version: int, resource_type: str) -> GetSourceCodeResponse:
    """
    Get the code for the given version and the given resource. The returned object
    includes constraints set at the project level that must be enforced on the agent
    during code install, if any.

    :param tid: The id of the environment
    :param version: The id of the model version
    :param resource_type: The type name of the resource
    :raises NotFound: Raised when the version or type is not found
    """


@typedmethod(
    path="/pip/config/<version>",
    operation="GET",
    api=True,
    agent_server=True,
    arg_options=methods.ENV_OPTS,
    client_types=[ClientType.agent, ClientType.api],
    api_version=2,
)
def get_pip_config(tid: uuid.UUID, version: int) -> Optional[model.PipConfig]:
    """
    Get the pip config for the given version

    :param tid: The id of the environment
    :param version: The id of the model version
    :raises NotFound: Raised when the version or environment is not found
    """


@typedmethod(
    path="/metrics",
    operation="GET",
    arg_options=methods.ENV_OPTS,
    client_types=[ClientType.api],
    api_version=2,
)
def get_environment_metrics(
    tid: uuid.UUID,
    metrics: Sequence[str],
    start_interval: datetime.datetime,
    end_interval: datetime.datetime,
    nb_datapoints: int,
    round_timestamps: bool = False,
) -> model.EnvironmentMetricsResult:
    """
    Obtain metrics about the given environment for the given time interval.

    :param tid: The id of the environment for which the metrics have to be collected.
    :param metrics: List of names of metrics that have to be returned.
    :param start_interval: The start of the time window for which the metrics should be returned.
    :param end_interval: The end of the time window for which the metrics should be returned.
    :param nb_datapoints: The amount of datapoint that will be returned within the given time interval for each metric.
    :param round_timestamps: If this parameter is set to True, the timestamps in the reply will be rounded to a full hour.
        All time windows in the reply will have an equal size. To achieve this the start_interval, end_interval and
        nb_datapoint in the reply may differ from the ones requested.

            * The start_interval may be smaller than requested
            * The end_interval may be larger than requested
            * The nb_datapoints may be larger than requested

    :raises BadRequest: start_interval >= end_interval
    :raises BadRequest: nb_datapoints < 0
    :raises BadRequest: The provided metrics list is an empty list.
    :raises BadRequest: The start_interval and end_interval are not separated from each other by at least nb_datapoints minutes
                        separated from each other.
    :raises BadRequest: The round_timestamps parameter is set to True and the amount of hours between
                        start_interval and end_interval is less than the requested number of datapoints.
    """


@typedmethod(path="/login", operation="POST", client_types=[ClientType.api], enforce_auth=False, api_version=2)
def login(username: str, password: str) -> ReturnValue[model.LoginReturn]:
    """Login a user.

     When the login succeeds an authentication header is returned with the Bearer token set.

    :param username: The user to login
    :param password: The password of this user
    :raises UnauthorizedException: Raised when the login failed or if server authentication is not enabled
    """


@typedmethod(path="/user", operation="GET", client_types=[ClientType.api], api_version=2)
def list_users() -> list[model.User]:
    """List all users

    :return: A list of all users"""


@typedmethod(path="/current_user", operation="GET", client_types=[ClientType.api], api_version=2)
def get_current_user() -> model.CurrentUser:
    """Get the current logged in user (based on the provided JWT) and server auth settings

    :raises NotFound: Raised when server authentication is not enabled
    """


@typedmethod(path="/user/<username>", operation="DELETE", client_types=[ClientType.api], api_version=2)
def delete_user(username: str) -> None:
    """Delete a user from the system with given username.

    :param username: The username to delete
    :raises NotFound: Raised when the user does not exist
    :raises BadRequest: Raised when server authentication is not enabled
    """


@typedmethod(path="/user", operation="POST", client_types=[ClientType.api], api_version=2)
def add_user(username: str, password: str) -> model.User:
    """Add a new user to the system

    :param username: The username of the new user. The username cannot be an empty string.
    :param password: The password of this new user. The password should be at least 8 characters long.
    :raises Conflict: Raised when there is already a user with this user_name
    :raises BadRequest: Raised when server authentication is not enabled
    """


@typedmethod(path="/user/<username>/password", operation="PATCH", client_types=[ClientType.api], api_version=2)
def set_password(username: str, password: str) -> None:
    """Change the password of a user

    :param username: The username of the user
    :param password: The password of this new user. The password should be at least 8 characters long.
    :raises NotFound: Raised when the user does not exist
    :raises BadRequest: Raised when server authentication is not enabled
    """


@typedmethod(
    path="/discovered/<discovered_resource_id>",
    operation="POST",
    agent_server=True,
    arg_options=methods.ENV_OPTS,
    client_types=[ClientType.agent],
    api_version=2,
    varkw=True,
)
def discovered_resource_create(
    tid: uuid.UUID,
    discovered_resource_id: str,
    discovery_resource_id: str,
    **kwargs: object,  # bypass the type checking for the values
) -> None:
    """
    create a discovered resource.

    :param tid: The id of the environment this resource belongs to
    :param discovered_resource_id: The id of the discovered_resource
    :param discovery_resource_id: The id of the discovery resource responsible for discovering this resource
    :param **kwargs: The following arguments are supported:
           values: The values associated with the discovered_resource
    """


@typedmethod(
    path="/discovered/",
    operation="POST",
    agent_server=True,
    arg_options=methods.ENV_OPTS,
    client_types=[ClientType.agent],
    api_version=2,
)
def discovered_resource_create_batch(tid: uuid.UUID, discovered_resources: Sequence[LinkedDiscoveredResource]) -> None:
    """
    create multiple discovered resource in the DB
    :param tid: The id of the environment this resource belongs to
    :param discovered_resources: List of discovered_resources containing the discovered_resource_id and values for each resource
    """


@typedmethod(
    path="/discovered/<discovered_resource_id>",
    operation="GET",
    arg_options=methods.ENV_OPTS,
    client_types=[ClientType.api],
    api_version=2,
)
def discovered_resources_get(tid: uuid.UUID, discovered_resource_id: ResourceIdStr) -> model.DiscoveredResource:
    """
    Get a single discovered resource.

    :param tid: the id of the environment in which to get the discovered resource.
    :param discovered_resource_id: The id of the discovered resource
    """


@typedmethod(
    path="/discovered",
    operation="GET",
    arg_options=methods.ENV_OPTS,
    client_types=[ClientType.api],
    api_version=2,
)
def discovered_resources_get_batch(
    tid: uuid.UUID,
    limit: Optional[int] = None,
    start: Optional[str] = None,
    end: Optional[str] = None,
    sort: str = "discovered_resource_id.asc",
    filter: Optional[Mapping[str, Sequence[str]]] = None,
) -> list[model.DiscoveredResource]:
    """
    Get a list of discovered resources.

    The discovery resource responsible for discovering each resource is included.
    For resources that the orchestrator is already managing, a link to the corresponding resource is provided. The full list of
    supported links can be found :ref:`here <api_self_referencing_links>`.


    :param tid: The id of the environment this resource belongs to
    :param limit: Limit the number of instances that are returned
    :param start: The lower limit for the order by column (exclusive).
                Only one of 'start' and 'end' should be specified at the same time.
    :param end: The upper limit for the order by column (exclusive).
                Only one of 'start' and 'end' should be specified at the same time.
    :param sort: Return the results sorted according to the parameter value.
            The following sorting attributes are supported: 'discovered_resource_id', 'resource_type', 'agent',
            'resource_id_value'.
            The following orders are supported: 'asc', 'desc'
    :param filter: Filter the list of returned resources.
        Default behavior: return all discovered resources.
        Filtering by 'managed' is supported:

            - filter.managed=true: only return discovered resources that the orchestrator is already aware of i.e.
              resources that are present in any released configuration model of environment tid.
            - filter.managed=false: only return discovered resources that the orchestrator is unaware of i.e. resources
              that are not part of any released configuration model of environment tid.
            - filter.resource_type: filter by resource type.
            - filter.agent: filter by agent.
            - filter.resource_id_value: filter by resource id value.

    :return: A list of all matching released resources
    :raise NotFound: This exception is raised when the referenced environment is not found
    :raise BadRequest: When the parameters used for filtering, sorting or paging are not valid
    """


@typedmethod(
    path="/discovered/<discovered_resource_id>",
    operation="DELETE",
    arg_options=methods.ENV_OPTS,
    client_types=[ClientType.api],
    api_version=2,
)
def discovered_resource_delete(
    tid: uuid.UUID,
    discovered_resource_id: str,
) -> None:
    """
    Delete a discovered resource.

    :param tid: The id of the environment this resource belongs to
    :param discovered_resource_id: The id of the discovered_resource

    :raise NotFound: When the referenced discovered resource is not found on the environment
    """


@typedmethod(
    path="/discovered/",
    operation="DELETE",
    arg_options=methods.ENV_OPTS,
    client_types=[ClientType.api],
    api_version=2,
)
def discovered_resource_delete_batch(tid: uuid.UUID, discovered_resource_ids: Sequence[str]) -> None:
    """
    Delete multiple discovered resources.
    If one or more discovered resources are not found on the environment,
    they will be ignored and the other discovered resources will be deleted.
    A 200 will be returned even if any/all provided discovered resources are ignored.

    :param tid: The id of the environment this resource belongs to
    :param discovered_resource_ids: List of discovered resource ids to delete
    """


@typedmethod(path="/health", operation="GET", client_types=[ClientType.api], enforce_auth=False, api_version=2)
def health() -> ReturnValue[None]:
    """
    Returns a 200 response code if the server is healthy
    or a 500 if the server is not healthy.

    In contrast to the 'GET /api/v1/serverstatus' endpoint, this endpoint does not require authentication.
    """
    pass
